"""
Admin Global api functions as a mixin
"""
from jaseci.api.interface import Interface
from jaseci.actor.sentinel import Sentinel
import uuid


class GlobalApi:
    """
    Admin global APIs
    """

    @Interface.admin_api(cli_args=["name"])
    def global_set(self, name: str, value: str):
        """
        Set a global
        """
        ret = {"success": True}
        if name == "GLOB_SENTINEL" or name in self._valid_configs:
            ret["response"] = f"{name} is sacred!"
            ret["success"] = False
        else:
            self._h.save_glob(name, value)
            ret["response"] = f"Global variable '{name}' to '{value}' set!"
        return ret

    @Interface.admin_api(cli_args=["name"])
    def global_delete(self, name: str):
        """
        Delete a global
        """
        ret = {"success": True}
        if name == "GLOB_SENTINEL" or name in self._valid_configs:
            ret["response"] = f"{name} is sacred!"
            ret["success"] = False
        else:
            self._h.destroy_glob(name)
            ret["response"] = f"Global {name} deleted."
        return ret

    @Interface.admin_api()
    def global_sentinel_set(self, snt: Sentinel = None):
        """
        Set sentinel as globally accessible
        """
        snt.make_read_only()
        snt.propagate_access()
        self._h.save_glob("GLOB_SENTINEL", snt.jid)
        return {"response": f"Global sentinel set to '{snt}'!"}

    @Interface.admin_api()
    def global_sentinel_unset(self):
        """
        Set sentinel as globally accessible
        """
        current = self.global_get("GLOB_SENTINEL")["value"]
        if current:
            snt = self._h.get_obj(self._m_id, uuid.UUID(current))
            for i in snt.get_deep_obj_list():
                i.make_private()
        self._h.destroy_glob("GLOB_SENTINEL")
        return {"response": "Global sentinel cleared!"}



from jaseci.jac.ir.passes.ir_pass import IrPass


class PrinterPass(IrPass):
    def __init__(self, *args):
        super().__init__(*args)

    def enter_node(self, node):
        print("Entering", node.name)

    def exit_node(self, node):
        print("Exiting", node.name)



from jaseci.utils.test_core import CoreTest


class InterpreterTest(CoreTest):
    """Unit tests for Jac Interpreter / Language features"""

    fixture_src = __file__

    def test_has_var_plucking(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("lang_features.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "has_var_plucking"}])
        self.assertEqual(
            ret["report"], [["node0", "node1", "node2"], ["edge0", "edge1", "edge2"]]
        )

    def test_deref_adaptive(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("lang_features.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "deref_adaptive"}])
        self.assertTrue(ret["report"][0].startswith("junk and stuff"))
        self.assertEqual(ret["report"][1], {"name": "node0"})

    def test_deref_of_element_fails(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("lang_features.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "deref_of_element_fails"}])
        self.assertFalse(ret["success"])



dot_node = """
    node test_node {has name;}
    graph test_graph {
        has anchor graph_root;
        graph G {
            graph_root [node=test_node, name=graph_root_node_name];
        }
    }
    walker init {
        root {
            spawn here --> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            std.out(here.name);
        }
    }
    """

dot_node_no_dot_id = """
    node test_node {has name;}
    graph test_graph {
        has anchor graph_root;
        graph {
            graph_root [node=test_node, name=graph_root_node_name];
        }
    }
    walker init {
        root {
            spawn here --> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            std.out(here.name);
        }
    }
    """

spawn_graph_node = """
    node test_node {has name;}
    graph test_graph {
        has anchor graph_root;
        spawn {
            graph_root =
                spawn node::test_node(name="graph_root_node_name");
        }
    }
    walker init {
        root {
            spawn here --> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            std.out(here.name);
        }
    }
    """

dot_node_multi_stmts = """
    node test_node {
        has name, date;
    }
    graph test_graph {
        has anchor graph_root;
        graph G {
            graph_root [node=test_node]
            graph_root [name=test_node, date=2021]
            graph_root [name=real_test_node]
        }
    }
    walker init {
        root {
            spawn here --> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            std.out(here.name);
            std.out(here.date);
        }
    }
    """

dot_edge = """
    node test_node {
        has name;
    }
    graph test_graph {
        has anchor graph_root;
        graph G {
            graph_root [node=test_node, name=root]
            node_1 [node=test_node, name=node_1]
            node_2 [node=test_node, name=node_2]
            graph_root -> node_1
            graph_root -> node_2
        }
    }
    walker init {
        has nodes;
        with entry {
            nodes = [];
        }
        root {
            spawn here --> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            nodes += [here];
            take --> node::test_node;
        }
        with exit {
            for i in nodes {
                std.out(i.name);
            }
        }
    }
    """

dot_edge_with_attrs = """
    node test_node {
        has name;
    }
    edge special;
    graph test_graph {
        has anchor graph_root;
        graph G {
            graph_root [node=test_node, name=root]
            node_1 [node=test_node, name=node_1]
            node_2 [node=test_node, name=node_2]
            graph_root -> node_1 [edge=special]
            graph_root -> node_2
        }
    }
    walker init {
        has nodes;
        with entry {
            nodes = [];
        }
        root {
            spawn here --> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            nodes += [here];
            take -[special]-> node::test_node;
        }
        with exit {
            for i in nodes {
                std.out(i.name);
            }
        }
    }
    """

dot_edge_with_attrs_vars = """
    node test_node {
        has name;
    }
    edge special {
        has name;
    }
    graph test_graph {
        has anchor graph_root;
        graph G {
            graph_root [node=test_node, name=root]
            node_1 [node=test_node, name=node_1]
            node_2 [node=test_node, name=node_2]
            graph_root -> node_1 [edge=special, name=edge_1]
            graph_root -> node_2
        }
    }
    walker init {
        has nodes;
        with entry {
            nodes = [];
        }
        root {
            spawn here --> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            for i in -[special]->.edge {
                std.out(i.name);
            }
            for i in -[special(name == "edge_1")]-> {
                std.out(i.name);
            }
        }
    }
    """

dot_graph = """
    node year {
        has color;
    }
    node month {
        has count, season;
    }
    node week;
    node day;
    edge parent;
    edge child;
    graph test_graph {
        has anchor A;
        strict graph G {
            H [node=year]
            C [node=week]
            E [node=day]
            D [node=day]

            A -> B // Basic directional edge
            B -- H // Basic non-directional edge
            B -> C [edge=parent] // Edge with attribute
            C -> D -> E [edge=child] // Chain edge

            A [color=red] // Node with DOT builtin graphing attr
            B [node=month, count=2] [season=spring]// Node with Jac attr
            A [node=year] // Multiple attr statement per node
        }
    }
    walker init {
        root {
            spawn here --> graph::test_graph;
        }
    }
    """

dot_quoted_string = """
    node test_node {
        has name;
    }
    graph test_graph {
        has anchor graph_root;
        graph G {
            graph_root [node=test_node, name=root]
            node_1 [node=test_node, name="this has space"]
            node_2 [node=test_node, name="another space"]
            graph_root -> node_1
            graph_root -> node_2
        }
    }
    walker init {
        root {
            spawn here --> graph::test_graph;
        }
        test_node {
            report here.name;
        }
        take -->;
    }
    """



"""
AST for Jac
"""

from jaseci.utils.utils import logger


class Ast:
    """
    AST Nodes
    """

    def __init__(
        self,
        mod_name,
    ):
        self.name = "unparsed"
        self.kid = []
        self.loc = [0, 0, mod_name if mod_name is not None else "@default", {}]

    def is_terminal(self):
        """Returns true if node is a terminal"""
        return len(self.loc[3].keys())

    def token(self):
        if not self.is_terminal():
            logger.error(str(f"Non terminals (rules) don't have token info - {self}"))
            return None
        else:
            return self.loc[3]["token"]

    def token_text(self):
        if self.is_terminal():
            return self.token()["text"]

    def token_symbol(self):
        if self.is_terminal():
            return self.token()["symbol"]

    def __str__(self):
        res = f"{self.name}:{self.loc[2]}:{self.loc[0]}:{self.loc[1]}:"
        if self.is_terminal():
            res += f':{self.loc[3]["token"]["text"]}'
        return res

    def __repr__(self):
        return self.__str__()

    def get_tokens(self):
        """Return list of all tokens derived from this ast node"""
        tokens = []
        if "token" in self.loc[3]:
            tokens.append(self.loc[3]["token"])
            return tokens
        else:
            for i in self.kid:
                tokens = tokens + i.get_tokens()
        return tokens

    def get_text(self):
        """Get source text for this node"""
        ret = ""
        for i in self.get_tokens():
            ret += f"{i['text']} "
        return ret



from json import dumps, loads
import sys
from jaseci.utils.utils import find_class_and_import


class MemoryHook:
    """
    Set of virtual functions to be used as hooks to allow access to
    the complete set of items across jaseci object types. This class contains
    a number of blank function calls to be bound extrenally and passed
    to the objects. They return jaseci core types.
    """

    def __init__(self):
        from jaseci.actions.live_actions import get_global_actions

        self.mem = {"global": {}}
        self._machine = None
        self.save_obj_list = set()
        self.save_glob_dict = {}
        self.global_action_list = get_global_actions(self)

    ####################################################
    #               COMMON GETTER/SETTER               #
    ####################################################

    # --------------------- OBJ ---------------------- #

    def get_obj(self, caller_id, item_id, override=False):
        """
        Get item from session cache by id, then try store
        TODO: May need to make this an object copy so you cant do mem writes
        """
        ret = self.get_obj_from_store(item_id)
        if override or (ret is not None and ret.check_read_access(caller_id)):
            return ret

    def has_obj(self, item_id):
        """
        Checks for object existance
        """
        return self.has_obj_in_store(item_id)

    def save_obj(self, caller_id, item, all_caches=False):
        """Save item to session cache, then to store"""
        if item.check_write_access(caller_id):
            self.commit_obj_to_cache(item, all_caches=all_caches)
            if item._persist:
                self.save_obj_list.add(item)

    def destroy_obj(self, caller_id, item):
        """Destroy item from session cache then  store"""
        if item.check_write_access(caller_id):
            self.decommit_obj_from_cache(item)
            if item._persist:
                self.destroy_obj_from_store(item)

    # --------------------- GLOB --------------------- #

    def get_glob(self, name):
        """
        Get global config from session cache by id, then try store
        """
        return self.get_glob_from_store(name)

    def has_glob(self, name):
        """
        Checks for global config existance
        """
        return self.has_glob_in_store(name)

    def resolve_glob(self, name, default=None):
        """
        Util function for returning config if exists otherwise default
        """
        if self.has_glob(name):
            return self.get_glob(name)
        else:
            return default

    def save_glob(self, name, value, persist=True):
        """Save global config to session cache, then to store"""
        self.commit_glob_to_cache(name, value)

        if persist:
            self.save_glob_dict[name] = value

    def list_glob(self):
        """Lists all configs present"""
        return self.list_glob_from_store()

    def destroy_glob(self, name, persist=True):
        """Destroy global config from session cache then store"""
        self.decommit_glob_from_cache(name)

        if persist:
            self.destroy_glob_from_store(name)

    # ----------------- GLOB CONFIG ------------------ #

    def build_config(self, name, config):
        if not self.has_glob(name):
            self.save_glob(name, dumps(config))
            self.commit()
        return loads(self.get_glob(name))

    ####################################################
    #        DATASOURCE METHOD (TO BE OVERRIDE)        #
    ####################################################

    # --------------------- OBJ ---------------------- #

    def get_obj_from_store(self, item_id):
        """
        Get item from externally hooked general store by id
        """
        if item_id.urn in self.mem:
            return self.mem[item_id.urn]

        return None

    def has_obj_in_store(self, item_id):
        """
        Checks for object existance in store
        """
        return item_id.urn in self.mem

    def destroy_obj_from_store(self, item):
        """Destroy item to externally hooked general store"""
        if item in self.save_obj_list:
            self.save_obj_list.remove(item)

    # --------------------- GLOB --------------------- #

    def get_glob_from_store(self, name):
        """
        Get global config from externally hooked general store by name
        """
        if name in self.mem["global"]:
            return self.mem["global"][name]

        return None

    def has_glob_in_store(self, name):
        """
        Checks for global config existance in store
        """
        return name in self.mem["global"]

    def destroy_glob_from_store(self, name):
        """Destroy global config to externally hooked general store"""
        if name in self.save_glob_dict:
            self.save_glob_dict.pop(name)

    def list_glob_from_store(self):
        """Get list of global config to externally hooked general store"""
        return list(self.mem["global"].keys())

    ####################################################
    # ------------------ COMMITTER ------------------- #
    ####################################################

    def commit(self):
        for i in self.save_obj_list:
            self.commit_obj_to_cache(i)

        self.save_obj_list = set()

        for i in self.save_glob_dict.keys():
            self.commit_glob_to_cache(name=i, value=self.save_glob_dict[i])

        self.save_glob_dict = {}

    ###################################################
    #   CACHE CONTROL (SHOULD NOT OVERRIDEN ON ORM)   #
    ###################################################

    # -------------------- GLOBS -------------------- #

    def commit_glob_to_cache(self, name, value):
        self.mem["global"][name] = value

    def decommit_glob_from_cache(self, name):
        self.mem["global"].pop(name)

    # --------------------- OBJ --------------------- #

    def has_id_in_mem_cache(self, id):
        return id is not None and id in self.mem

    def commit_obj_to_cache(self, item, all_caches=False):
        self.mem[item.id.urn] = item

    def commit_all_cache_sync(self):
        for i in self.save_obj_list:
            self.commit_obj_to_cache(i, all_caches=True)

    def decommit_obj_from_cache(self, item):
        self.mem.pop(item.id.urn)

    ####################################################
    # ------------------ UTILITIES ------------------- #
    ####################################################

    def get_object_distribution(self):
        dist = {}
        for i in self.mem.keys():
            t = type(self.mem[i])
            if t in dist.keys():
                dist[t] += 1
            else:
                dist[t] = 1
        return dist

    def mem_size(self):
        return sys.getsizeof(self.mem) / 1024

    ###################################################
    #                  CLASS CONTROL                  #
    ###################################################

    def find_class_and_import(self, j_type, mod):

        cls = getattr(self.meta, j_type, False)

        if not cls:
            cls = find_class_and_import(j_type, mod)

        return cls

    def clear_cache(self):
        MemoryHook.__init__(self)



"""Built in actions for Jaseci"""
from jaseci.actions.live_actions import jaseci_action


@jaseci_action()
def get_result(task_id, wait, meta):
    """
    Get task result by task_id
    """
    task = meta["h"].task

    if not task.is_running():
        raise Exception("Task hook is not yet initialized!")

    return task.get_by_task_id(task_id, wait)



"""
Sentinel interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.actor.architype import Architype
from jaseci.jac.interpreter.interp import Interp
from jaseci.utils.utils import parse_str_token
from jaseci.jac.ir.jac_code import jac_ast_to_ir
from jaseci.jac.machine.jac_scope import JacScope


class SentinelInterp(Interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    def run_start(self, jac_ast):
        """
        start: ver_label? element* EOF;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "ver_label":
            self.run_ver_label(kid[0])
        for i in kid[:-1]:
            self.run_element(i)

    def run_ver_label(self, jac_ast):
        """
        ver_label: 'version' COLON STRING;
        """
        kid = self.set_cur_ast(jac_ast)
        self.version = parse_str_token(kid[2].token_text())

    def run_element(self, jac_ast):
        """
        element: global_var | architype | test;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "global_var":
            self.load_global_var(kid[0])
        elif kid[0].name == "architype":
            self.load_architype(kid[0])
        elif kid[0].name == "test":
            self.load_test(kid[0])

    def load_global_var(self, jac_ast):
        """
        global_var:
            KW_GLOBAL NAME EQ expression (COMMA NAME EQ expression)* SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        kid = kid[1:]
        while kid[0].name == "NAME":
            const_name = kid[0].token_text()
            if const_name in self.global_vars:
                self.rt_error(f"Global {const_name} already defined!", kid[0])
            else:
                self.global_vars[const_name] = self.run_expression(kid[2]).value
            kid = kid[4:] if kid[3].name == "COMMA" else kid[3:]

    def load_architype(self, jac_ast):
        """
        architype:
            KW_NODE NAME (COLON NAME)* (COLON INT)? attr_block
            | KW_EDGE NAME (COLON NAME)* attr_block
            | KW_GRAPH NAME graph_block
            | KW_ASYNC? KW_WALKER NAME namespaces? walker_block;
        """
        kid = self.set_cur_ast(jac_ast)

        is_async = kid[0].name == "KW_ASYNC" and bool(kid.pop(0))

        name = kid[1].token_text()
        kind = kid[0].token_text()

        arch = Architype(
            m_id=self._m_id,
            h=self._h,
            code_ir=jac_ast,
            name=name,
            kind=kind,
            is_async=is_async,
            parent=self,
        )

        if len(kid) > 2 and kid[2].name == "COLON":
            for i in kid[2:]:
                if i.name == "NAME":
                    arch.super_archs.append(i.token_text())
        if self.arch_ids.has_obj_by_name(arch.name, kind=arch.kind):
            self.arch_ids.destroy_obj_by_name(arch.name, kind=arch.kind)
        self.arch_ids.add_obj(arch)
        self.arch_has_preproc(kid[-1], arch)
        self.arch_can_compile(kid[-1], arch)
        return arch

    # Note: Sentinels only registers the attr_stmts

    def arch_has_preproc(self, jac_ast, arch):
        """Helper function to statically compile can stmts for arch"""
        kid = self.set_cur_ast(jac_ast)
        if jac_ast.name in ["attr_block", "walker_block"]:
            for i in kid:
                if i.name == "attr_stmt" and i.kid[0].name == "has_stmt":
                    for j in i.kid[0].kid:
                        if j.name == "has_assign":
                            has_kid = j.kid
                            is_private = False
                            is_anchor = False
                            if has_kid[0].name == "KW_PRIVATE":
                                has_kid = has_kid[1:]
                                is_private = True
                            if has_kid[0].name == "KW_ANCHOR":
                                has_kid = has_kid[1:]
                                is_anchor = True
                            var_name = has_kid[0].token_text()
                            if is_private:
                                arch.private_vars.append(var_name)
                            if is_anchor and arch.anchor_var is None:
                                arch.anchor_var = var_name
                            arch.has_vars.append(var_name)
        arch.save()

    def arch_can_compile(self, jac_ast, arch):
        """Helper function to statically compile can stmts for arch"""
        kid = self.set_cur_ast(jac_ast)
        self.push_scope(JacScope(parent=self, has_obj=self, action_sets=[]))
        if jac_ast.name in ["attr_block", "walker_block"]:
            for i in kid:
                if i.name == "attr_stmt" and i.kid[0].name == "can_stmt":
                    self.run_can_stmt(i.kid[0], arch)
        elif kid[0].name == "graph_block_spawn":
            kid = kid[0].kid[2].kid
            for i in kid:
                self.run_can_stmt(i, arch)
        self.pop_scope()

    def load_test(self, jac_ast):
        """
        test:
            KW_TEST STRING KW_WITH (graph_ref | KW_GRAPH graph_block) KW_BY (
                (walker_ref spawn_ctx? (code_block | SEMI))
                | KW_WALKER walker_block
            );
        """
        kid = self.set_cur_ast(jac_ast)
        testcase = {
            "title": kid[1].token_text(),
            "graph_ref": None,
            "graph_block": None,
            "walker_ref": None,
            "spawn_ctx": None,
            "assert_block": None,
            "walker_block": None,
            "outcome": None,
        }
        kid = kid[3:]
        if kid[0].name == "graph_ref":
            graph_name = kid[0].kid[2].token_text()
            if not self.arch_ids.has_obj_by_name(graph_name, kind="graph"):
                self.rt_error(f"Graph {graph_name} not found!", kid[0])
                return
            testcase["graph_ref"] = graph_name
        else:
            kid = kid[1:]
            testcase["graph_block"] = jac_ast_to_ir(kid[0])
        kid = kid[2:]
        if kid[0].name == "walker_ref":
            walker_name = kid[0].kid[2].token_text()
            if not self.arch_ids.has_obj_by_name(name=walker_name, kind="walker"):
                self.rt_error(f"Walker {walker_name} not found!", kid[0])
                return
            testcase["walker_ref"] = walker_name
            kid = kid[1:]
            if kid[0].name == "spawn_ctx":
                testcase["spawn_ctx"] = jac_ast_to_ir(kid[0])
                kid = kid[1:]
            if kid[0].name == "code_block":
                testcase["assert_block"] = jac_ast_to_ir(kid[0])
        else:
            kid = kid[1:]
            testcase["walker_block"] = jac_ast_to_ir(kid[0])

        self.testcases.append(testcase)



from .state import ServiceState as Ss


class CommonService:
    def __init__(self, cls):
        self.cls = cls
        if not hasattr(self.cls, "_app"):
            setattr(self.cls, "_app", None)
            setattr(self.cls, "_state", Ss.NOT_STARTED)
            setattr(self.cls, "_quiet", True)

    @property
    def app(self):
        return self.cls._app

    @app.setter
    def app(self, val):
        self.cls._app = val

    @property
    def state(self) -> Ss:
        return self.cls._state

    @state.setter
    def state(self, val: Ss):
        self.cls._state = val

    @property
    def quiet(self) -> bool:
        return self.cls._quiet

    @quiet.setter
    def quiet(self, val: bool):
        self.cls._quiet = val

    ###################################################
    #                     COMMONS                     #
    ###################################################

    def is_ready(self):
        return self.state.is_ready() and self.app is None

    def is_running(self):
        return self.state.is_running() and not (self.app is None)

    def has_failed(self):
        return self.state.has_failed()

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def build(self, hook):
        self.app = None
        self.state = Ss.NOT_STARTED
        self.__init__(hook)


class ProxyService(CommonService):
    def __init__(self):
        super().__init__(ProxyService)


class MetaProperties:
    def __init__(self, prop):
        if not hasattr(prop, "_services"):
            setattr(prop, "_services", {})
            setattr(prop, "_background", {})
            setattr(prop, "_hook", None)
            setattr(prop, "_hook_param", {})
            setattr(prop, "_master", None)
            setattr(prop, "_super_master", None)

    @property
    def services(self) -> dict:
        return self.cls._services

    @services.setter
    def services(self, val: dict):
        self.cls._services = val

    @property
    def background(self) -> dict:
        return self.cls._background

    @background.setter
    def background(self, val: dict):
        self.cls._background = val

    @property
    def hook(self):
        return self.cls._hook

    @hook.setter
    def hook(self, val):
        self.cls._hook = val

    @property
    def hook_param(self) -> dict:
        return self.cls._hook_param

    @hook_param.setter
    def hook_param(self, val: dict):
        self.cls._hook_param = val

    @property
    def master(self):
        return self.cls._master

    @master.setter
    def master(self, val):
        self.cls._master = val

    @property
    def super_master(self):
        return self.cls._super_master

    @super_master.setter
    def super_master(self, val):
        self.cls._super_master = val



from copy import copy
from unittest.mock import MagicMock, Mock

from jaseci.svc import MailService
from jaseci.svc.mail import MAIL_CONFIG
from jaseci.utils.test_core import CoreTest


class MailLibTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def __init__(self, *args, **kwargs):
        MailService.connect = MagicMock(return_value=Mock())
        super(MailLibTest, self).__init__(*args, **kwargs)

    def test_send_mail(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("mail_test.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "send_mail"}])
        self.assertTrue(ret["success"])

        ms = MailService()
        configs = copy(MAIL_CONFIG)
        configs.pop("quiet")

        self.assertEqual(ms.connect.call_args[0], (configs,))

        self.assertEqual(
            ms.app.method_calls[0].args,
            (None, ["jaseci.dev@gmail.com"], "Test Subject", ("Test", "<h1>Test</h1>")),
        )



"""
Super master handler for each user of Jaseci, serves as main interface between
between user and Jaseci
"""

from jaseci.element.master import Master
from jaseci.api.logger_api import LoggerApi
from jaseci.api.config_api import ConfigApi
from jaseci.api.global_api import GlobalApi
from jaseci.api.super_api import SuperApi
from jaseci.api.actions_api import ActionsApi


class SuperMaster(Master, LoggerApi, ActionsApi, ConfigApi, GlobalApi, SuperApi):
    """Master with admin APIs"""

    def __init__(self, *args, **kwargs):
        Master.__init__(self, *args, **kwargs)



from jaseci.utils.test_core import CoreTest
import jaseci.tests.jac_test_code as jtc


class GlobalApiTest(CoreTest):
    """Unit tests for Jac Global APIs"""

    fixture_src = __file__

    def setUp(self):
        super().setUp()

        self.smast2 = self.meta.build_super_master(h=self.smast._h)
        self.smast.sentinel_register(name="test", code=jtc.basic)

        self.mast = self.meta.build_master()
        self.mast2 = self.meta.build_super_master(h=self.mast._h)
        self.mast.sentinel_register(name="test", code=jtc.basic)

    def tearDown(self):
        super().tearDown()

    def test_sent_global_active(self):
        """Test setting global sentinel"""
        api = ["global_sentinel_set", {"snt": None}]
        self.call(self.smast, api)
        api = ["sentinel_active_global", {}]
        self.call(self.smast2, api)
        api = ["sentinel_list", {"detailed": False}]
        r = self.call(self.smast2, api)
        self.assertEqual(len(r), 0)
        api = ["sentinel_active_get", {"detailed": False}]
        r = self.call(self.smast2, api)
        self.assertIn("jid", r.keys())
        self.assertEqual(r["name"], "test")
        api = ["sentinel_list", {"detailed": False}]
        r = self.call(self.smast2, api)
        self.assertEqual(len(r), 0)

    def test_sent_global_pull(self):
        """Test setting global sentinel"""
        api = ["global_sentinel_set", {"snt": None}]
        self.call(self.smast, api)
        api = ["sentinel_list", {"detailed": False}]
        r = self.call(self.smast2, api)
        self.assertEqual(len(r), 0)
        api = ["sentinel_pull", {}]
        self.call(self.smast2, api)
        api = ["sentinel_list", {"detailed": False}]
        r = self.call(self.smast2, api)
        self.assertEqual(len(r), 1)
        api = ["sentinel_active_get", {"detailed": False}]
        r = self.call(self.smast2, api)
        self.assertIn("jid", r.keys())
        self.assertEqual(r["name"], "test")

    def test_global_set_get_delete(self):
        """Test setting global sentinel"""
        api = ["global_get", {"name": "apple"}]
        r = self.call(self.smast, api)
        self.assertIsNone(r["value"])
        api = ["global_set", {"name": "apple", "value": "56"}]
        r = self.call(self.smast, api)
        api = ["global_get", {"name": "apple"}]
        r = self.call(self.smast2, api)
        self.assertEqual(r["value"], "56")
        api = ["global_delete", {"name": "apple"}]
        r = self.call(self.smast2, api)
        api = ["global_get", {"name": "apple"}]
        r = self.call(self.smast, api)
        self.assertIsNone(r["value"])

    def test_user_create(self):
        """Test master create operation"""
        api = ["user_create", {"name": "yo@gmail.com"}]
        r = self.call(self.mast, api)
        self.assertIn("j_type", r["user"])
        self.assertEqual(r["user"]["j_type"], "master")

    def test_master_create(self):
        """Test master create operation"""
        api = ["master_create", {"name": "yo@gmail.com"}]
        r = self.call(self.mast, api)
        self.assertIn("j_type", r["user"])
        self.assertEqual(r["user"]["j_type"], "master")

    def test_master_create_error_out(self):
        """Test master create operation"""
        api = ["master_create", {"name": "yo@gmail.com"}]
        r = self.call(self.mast, api)
        api = ["master_create", {"name": "yo@gmail.com"}]
        r = self.call(self.mast, api)
        self.assertIn("response", r)
        self.assertIn("already exists", r["response"])

    def test_master_create_super_limited(self):
        """Test master create operation"""
        api = ["master_createsuper", {"name": "yo3@gmail.com"}]
        r = self.call(self.mast, api)
        self.assertIn("response", r)
        self.assertIn("not a valid", r["response"])

    def test_master_create_linked_super_master_create(self):
        """Test master create operation"""
        api = ["master_createsuper", {"name": "yo3@gmail.com"}]
        r = self.call(self.smast, api)
        self.assertIn("j_type", r["user"])
        self.assertEqual(r["user"]["j_type"], "super_master")

    def test_global_sentinel_set_unset(self):
        api = ["global_sentinel_set", {}]
        r = self.call(self.smast, api)
        self.assertIn("response", r)
        self.assertNotIn("error", r)
        api = ["global_sentinel_unset", {}]
        r = self.call(self.smast, api)
        self.assertIn("response", r)
        self.assertNotIn("error", r)

    def test_global_sentinel_double_unset(self):
        api = ["global_sentinel_set", {}]
        r = self.call(self.smast, api)
        self.assertIn("response", r)
        self.assertNotIn("error", r)
        api = ["global_sentinel_unset", {}]
        r = self.call(self.smast, api)
        self.assertIn("response", r)
        self.assertNotIn("error", r)
        api = ["global_sentinel_unset", {}]
        r = self.call(self.smast, api)
        self.assertIn("response", r)
        self.assertNotIn("error", r)

    def test_graph_node_set_mem_leak_fix(self):
        api = ["graph_create", {}]
        r = self.call(self.smast, api)
        before = len(self.smast._h.mem)
        jid = r["jid"]
        api = ["graph_node_set", {"nd": jid, "ctx": {}}]
        r = self.call(self.smast, api)
        api = ["graph_node_set", {"nd": jid, "ctx": {}}]
        r = self.call(self.smast, api)
        api = ["graph_node_set", {"nd": jid, "ctx": {}}]
        r = self.call(self.smast, api)
        api = ["graph_node_set", {"nd": jid, "ctx": {}}]
        r = self.call(self.smast, api)
        api = ["graph_node_set", {"nd": jid, "ctx": {}}]
        r = self.call(self.smast, api)
        api = ["graph_node_set", {"nd": jid, "ctx": {}}]
        r = self.call(self.smast, api)
        after = len(self.smast._h.mem)
        self.assertEqual(before, after)

    def test_mem_no_leak_creating_one_node_and_node_set(self):
        mem_state = self.smast._h.mem
        before = len(mem_state)
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("hello_world.jac")}],
        )
        self.assertEqual(len(mem_state) - before, 8)
        before = len(mem_state)
        r = self.call(
            self.smast,
            ["walker_run", {"name": "one_node"}],
        )
        jid = r["report"][0]["jid"]
        self.assertEqual(len(mem_state) - before, 2)
        before = len(mem_state)
        self.call(
            self.smast,
            ["graph_node_set", {"nd": jid, "ctx": {"msg": "goodbye world"}}],
        )
        self.assertEqual(len(mem_state) - before, 0)
        r = self.call(
            self.smast,
            ["graph_node_get", {"nd": jid}],
        )
        self.assertEqual(r["msg"], "goodbye world")



"""
Sentinel api functions as a mixin
"""
from jaseci.api.interface import Interface
from jaseci.utils.id_list import IdList
from jaseci.actor.sentinel import Sentinel
from jaseci.utils.utils import b64decode_str
import uuid


class SentinelApi:
    """
    Sentinel APIs

    A sentinel is a unit in Jaseci that represents the organization and management of
    a collection of architypes and walkers. In a sense, you can think of a sentinel
    as a complete Jac implementation of a program or API application. Though its the
    case that many sentinels can be interchangeably across any set of graphs, most
    use cases will typically be a single sentinel shared by all users and managed by an
    admin(s), or each users maintaining a single sentinel customized for their
    individual needs. Many novel usage models are possible, but I'd point the beginner
    to the model most analogous to typical server side software development to start
    with. This model would be to have a single admin account responsible for updating
    a single sentinel that all users would share for their individual graphs. This
    model is achieved through using \\texttt{sentinel_register},
    \\texttt{sentinel_active_global}, and \\texttt{global_sentinel_set}.
    """

    def __init__(self):
        self.active_snt_id = None
        self.sentinel_ids = IdList(self)

    @Interface.private_api(cli_args=["code"])
    def sentinel_register(
        self,
        name: str = "default",
        code: str = "",
        code_dir: str = "./",
        mode: str = "default",
        encoded: bool = False,
        auto_run: str = "init",
        auto_run_ctx: dict = {},
        auto_create_graph: bool = True,
        set_active: bool = True,
    ):
        """
        Create blank or code loaded sentinel and return object
        Auto_run is the walker to execute on register (assumes active graph
        is selected)
        """
        snt = self.sentinel_ids.get_obj_by_name(name, silent=True)
        new_gph = None

        if not snt:
            snt = Sentinel(m_id=self._m_id, h=self._h, name=name)
            self.sentinel_ids.add_obj(snt)
            if auto_create_graph:
                new_gph = self.graph_create(set_active=set_active)
        if code:
            self.sentinel_set(
                code=code, code_dir=code_dir, encoded=encoded, snt=snt, mode=mode
            )
            if not snt.is_active:
                return {
                    "response": "Error in jac code",
                    "errors": snt.errors + snt.runtime_errors,
                    "success": False,
                }
        self.attempt_auto_run(sent=snt, walk_name=auto_run, ctx=auto_run_ctx)
        if set_active:
            self.sentinel_active_set(snt)
        self.extract_snt_aliases(snt)
        if new_gph:
            return [snt.serialize(), new_gph]
        return [snt.serialize()]

    @Interface.private_api()
    def sentinel_pull(self, set_active: bool = True, on_demand: bool = True):
        """
        Copies global sentinel to local master
        """
        glob_id = self._h.get_glob("GLOB_SENTINEL")
        if not glob_id:
            return {"response": "No global sentinel is available!", "success": False}
        g_snt = self._h.get_obj(self._m_id, uuid.UUID(glob_id)).duplicate()

        snt = self.sentinel_ids.get_obj_by_name(g_snt.name, silent=True)
        if not snt:
            snt = Sentinel(m_id=self._m_id, h=self._h, name=g_snt.name)
            self.sentinel_ids.add_obj(snt)
        elif on_demand and snt.is_active:
            return {"response": f"{snt} already active!", "success": True}
        if set_active:
            self.sentinel_active_set(snt)
        return self.sentinel_set(code=g_snt.code_ir, snt=snt, mode="ir")

    @Interface.private_api()
    def sentinel_get(
        self, snt: Sentinel = None, mode: str = "default", detailed: bool = False
    ):
        """
        Get a sentinel rendered with specific mode
        Valid modes: {default, code, ir, }
        """
        if mode == "code":
            return snt._jac_ast.get_text()
        elif mode == "ir":
            return snt.ir_dict()
        else:
            return snt.serialize(detailed=detailed)

    @Interface.private_api(cli_args=["code"])
    def sentinel_set(
        self,
        code: str,
        code_dir: str = "./",
        encoded: bool = False,
        snt: Sentinel = None,
        mode: str = "default",
    ):
        """
        Set code/ir for a sentinel, only replaces walkers/archs in sentinel
        Valid modes: {code, ir, }
        """
        if encoded:
            code = b64decode_str(code)
        if mode not in ["code", "default", "ir"]:
            return {"response": f"Invalid mode to set {snt}", "success": False}
        snt.register_code(code, dir=code_dir, mode=mode)
        snt.propagate_access()

        if snt.is_active:
            self.extract_snt_aliases(snt)
            return {"response": f"{snt} registered and active!", "success": True}
        else:
            return {
                "response": f"{snt} code issues encountered!",
                "success": False,
                "errors": snt.errors + snt.runtime_errors,
            }

    @Interface.private_api()
    def sentinel_list(self, detailed: bool = False):
        """
        Provide complete list of all sentinel objects
        """
        snts = []
        for i in self.sentinel_ids.obj_list():
            snts.append(i.serialize(detailed=detailed))
        return snts

    @Interface.private_api()
    def sentinel_test(self, snt: Sentinel = None, detailed: bool = False):
        """
        Run battery of test cases within sentinel and provide result
        """
        return snt.run_tests(detailed=detailed)

    @Interface.private_api(cli_args=["snt"])
    def sentinel_active_set(self, snt: Sentinel):
        """
        Sets the default sentinel master should use
        """
        self.active_snt_id = snt.jid
        self.alias_register("active:sentinel", snt.jid)
        return [f"Sentinel {snt.id} set as default"]

    @Interface.private_api()
    def sentinel_active_unset(self):
        """
        Unsets the default sentinel master should use
        """
        self.active_snt_id = None
        self.alias_delete("active:sentinel")
        return ["Default sentinel unset"]

    @Interface.private_api()
    def sentinel_active_global(
        self,
        auto_run: str = "",
        auto_run_ctx: dict = {},
        auto_create_graph: bool = False,
        detailed: bool = False,
    ):
        """
        Sets the default master sentinel to the global sentinel
        Exclusive OR with pull strategy
        """
        ret = {"success": False, "sentinel": None}
        glob_id = self._h.get_glob("GLOB_SENTINEL")
        if not glob_id:
            ret["response"] = "No global sentinel is available!"
        else:
            self.active_snt_id = "global"  # Resolved in interface
            self.alias_register("active:sentinel", glob_id)
            sent = self._h.get_obj(self._m_id, uuid.UUID(glob_id))
            if auto_create_graph:
                ret["graph_created"] = self.graph_create(set_active=True)
            auto_run_ret = self.attempt_auto_run(
                sent=sent, walk_name=auto_run, ctx=auto_run_ctx
            )
            if auto_run_ret:
                ret["auto_run_result"] = auto_run_ret
            ret["sentinel"] = sent.serialize(detailed=detailed)
            ret["success"] = True
            ret["response"] = f"Global sentinel {glob_id} set as default"
        return ret

    @Interface.private_api()
    def sentinel_active_get(self, detailed: bool = False):
        """
        Returns the default sentinel master is using
        """
        id = self.active_snt_id
        if id == "global":
            id = self._h.get_glob("GLOB_SENTINEL")
        if not id:
            return {"response": "No default sentinel is selected!", "success": False}
        else:
            default = self._h.get_obj(self._m_id, uuid.UUID(id))
            return default.serialize(detailed=detailed)

    @Interface.private_api(cli_args=["snt"])
    def sentinel_delete(self, snt: Sentinel):
        """
        Permanently delete sentinel with given id
        """
        self.remove_snt_aliases(snt)
        if self.active_snt_id == snt.jid:
            self.sentinel_active_unset()
        self.sentinel_ids.destroy_obj(snt)
        return [f"Sentinel {snt.id} successfully deleted"]

    def active_snt(self):
        return self._h.get_obj(self._m_id, uuid.UUID(self.active_snt_id))

    def attempt_auto_run(self, sent: Sentinel, walk_name, ctx):
        if (
            sent.arch_ids.has_obj_by_name(walk_name, kind="walker")
            and self.active_gph_id
        ):
            nd = self._h.get_obj(self._m_id, uuid.UUID(self.active_gph_id))
            return self.walker_run(name=walk_name, nd=nd, ctx=ctx, snt=sent)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.sentinel_ids.obj_list():
            i.destroy()



from jaseci.utils.test_core import CoreTest


class GraphApiTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_graph_depth(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("fam.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "create_fam"}])
        ret = self.call(self.mast, ["graph_get", {}])
        self.assertEqual(len(ret), 9)
        ret = self.call(self.mast, ["graph_get", {"depth": 2}])
        self.assertEqual(len(ret), 6)
        ret = self.call(self.mast, ["graph_get", {"depth": 1}])
        self.assertEqual(len(ret), 1)

    def test_graph_node_view(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("fam.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "create_fam"}])
        ret = self.call(self.mast, ["graph_get", {}])
        self.assertEqual(len(ret), 9)
        ret = self.call(self.mast, ["graph_node_view", {"show_edges": True}])
        self.assertEqual(len(ret), 5)
        ret = self.call(self.mast, ["graph_node_view", {}])
        self.assertEqual(len(ret), 3)

    def test_graph_node_view_filters(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("fam.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "create_fam"}])
        ret = self.call(self.mast, ["graph_node_view", {"node_type": "man"}])
        self.assertEqual(len(ret), 2)
        jid = ret[1]["jid"]
        ret = self.call(self.mast, ["graph_node_view", {"nd": jid, "show_edges": True}])
        self.assertEqual(len(ret), 7)
        ret = self.call(
            self.mast,
            [
                "graph_node_view",
                {"nd": jid, "show_edges": False, "edge_type": "married"},
            ],
        )
        self.assertEqual(len(ret), 2)
        ret = self.call(
            self.mast,
            [
                "graph_node_view",
                {"nd": jid, "show_edges": True, "edge_type": "married"},
            ],
        )
        self.assertEqual(len(ret), 3)



from jaseci.utils.test_core import CoreTest
import uuid


class StackTests(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_walker_yield(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("ll.jac")}],
        )
        self.call(self.mast, ["walker_run", {"name": "init"}])
        self.call(self.mast, ["walker_run", {"name": "gen_rand_life"}])
        ret = self.call(self.mast, ["walker_run", {"name": "get_gen_day"}])
        self.assertGreater(len(ret["report"]), 3)
        self.call(self.mast, ["walker_run", {"name": "get_gen_day"}])
        self.mast._h.commit()
        self.call(self.mast, ["walker_run", {"name": "get_gen_day"}])
        self.assertEqual(len(self.mast._h.save_obj_list), 0)

    def test_walker_context_auto_refresh(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("ll.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "print_life_note"}])
        life_node = self.mast._h.get_obj(self.mast._m_id, uuid.UUID(ret["final_node"]))
        life_node.context.pop("note")
        ret = self.call(self.mast, ["walker_run", {"name": "print_life_note"}])
        self.assertTrue(ret["success"])



import os
from unittest import TestCase

from antlr4 import CommonTokenStream, InputStream

import jaseci.actions.live_actions as lact
import jaseci.tests.jac_test_code as jtc
from jaseci.actor.sentinel import Sentinel
from jaseci.graph.graph import Graph
from jaseci.jac.jac_parse.jacLexer import jacLexer
from jaseci.jac.jac_parse.jacParser import jacParser
from jaseci.svc import MetaService
from jaseci.utils.utils import TestCaseHelper


class JacTests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()
        self.meta = MetaService()

    def tearDown(self):
        super().tearDown()

    def test_antlr4_parsing_lifelogify(self):
        """Basic test of jac grammar with lifelogify program"""
        input_stream = InputStream(jtc.prog1)
        lexer = jacLexer(input_stream)
        stream = CommonTokenStream(lexer)
        parser = jacParser(stream)
        parser.start()
        self.assertEqual(parser.getNumberOfSyntaxErrors(), 0)

    def test_sentinel_loading_jac_code(self):
        """Test the generation of jaseci trees for programs in grammar"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.prog1)
        self.assertIsNotNone(sent.run_architype("get_gen_day"))
        self.assertIsNotNone(sent.arch_ids.get_obj_by_name("week", kind="node"))

    def test_sentinel_loading_jac_code_multiple_times(self):
        """Test registering resets correctly for multiple attempts"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.prog0)
        num_arch = len(sent.arch_ids)
        sent.register_code(jtc.prog0)
        self.assertEqual(len(sent.arch_ids), num_arch)
        sent.register_code(jtc.prog0)
        self.assertEqual(len(sent.arch_ids), num_arch)

    def test_sentinel_register_dep_on_static_errors(self):
        """Test Jac registering is dependant on correct static/dynamic code"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        correct = "node b { has anchor a; }"
        wrong1 = "adfdsf"
        sent.register_code(correct)
        self.assertTrue(sent.is_active)
        sent.register_code(wrong1)
        self.assertFalse(sent.is_active)
        sent.register_code(correct)
        self.assertTrue(sent.is_active)
        sent.register_code(wrong1)
        self.assertFalse(sent.is_active)
        sent.register_code(correct)
        self.assertTrue(sent.is_active)

    def test_sentinel_loading_arhitype(self):
        """Test the generation of jaseci trees for programs in grammar"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.prog1)
        self.assertGreater(
            len(sent.arch_ids.get_obj_by_name("month", kind="node").code_ir), 5
        )

    def test_sentinel_running_basic_walker(self):
        """Test the execution of a basic walker building graph"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name("life", kind="node").run()
        test_walker = sent.run_architype("get_gen_day")
        test_walker.prime(test_node)
        test_walker.context["date"] = "2010-08-03T03:00:00.000000"
        test_walker.run()
        self.assertEqual(len(test_node.outbound_nodes()), 1)
        self.assertEqual(test_node.outbound_nodes()[0].name, "year")
        self.assertEqual(
            test_node.outbound_nodes()[0].outbound_nodes()[0].outbound_nodes()[0].name,
            "week",
        )

    def test_sentinel_setp_running_walker(self):
        """Test the execution of a basic walker building graph"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name("life", kind="node").run()
        test_walker = sent.run_architype("get_gen_day")
        test_walker.prime(test_node)
        test_walker.context["date"] = "2010-08-03T03:00:00.000000"
        self.assertEqual(len(test_node.outbound_nodes()), 0)
        next = test_walker.step()
        self.assertEqual(test_node.outbound_nodes()[0], next)
        next = test_walker.step()
        next = test_walker.step()
        self.assertEqual(next.name, "week")
        next = test_walker.step()
        next = test_walker.step()
        self.assertEqual(test_walker.current_step, 5)
        self.assertTrue(next)
        next = test_walker.step()
        self.assertFalse(next)

    def test_walker_writes_through_no_data_gain_loss_on_contexts(self):
        """
        Test that  no loss or gain of data on second trak on second trek
        """
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name("life", kind="node").run()
        test_walker = sent.run_architype("get_gen_day")
        test_walker.prime(test_node)
        test_walker.context["date"] = "2010-08-03T03:00:00.000000"
        self.assertEqual(test_walker.context["date"], "2010-08-03T03:00:00.000000")
        test_walker.run()
        after_gen = len(gph._h.mem)
        test_walker.prime(test_node)
        test_walker.context["date"] = "2010-08-03T03:00:00.000000"
        test_walker.run()
        after_track = len(gph._h.mem)
        # Tests that more items were created the first time through
        self.assertEqual(after_gen, after_track)

    def test_arch_ids_generate_node_bound_objects_scalably(self):
        """
        Test that arch_ids in sents bind contents to actual nodes
        scalably (node contexts dont get deleted when arch_ids deleted)
        """
        lact.load_local_actions(os.path.dirname(__file__) + "/infer.py")
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name("life", kind="node").run()
        test_walker = sent.run_architype("get_gen_day")
        test_walker.prime(test_node)
        test_walker.context["date"] = "2010-08-03T03:00:00.000000"
        self.assertEqual(test_walker.context["date"], "2010-08-03T03:00:00.000000")
        test_walker.run()

        year_node = test_node.outbound_nodes()[0]
        before_del = year_node.context["year"]
        sent.register_code(jtc.prog1)
        after_del = year_node.context["year"]
        self.assertEqual(before_del, after_del)

    def test_sent_loads_complex_walker_and_arch(self):
        """Test loading attributes of arch and walkers"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.prog1)
        test_node = sent.arch_ids.get_obj_by_name("testnode", kind="node").run()
        test_walker = sent.run_architype("testwalk")
        test_walker.prime(test_node)
        test_walker.run()
        self.assertEqual(test_node.context["c"], '43Yeah \n"fools"!')

    def test_availabilty_of_global_functions(self):
        """Test preset function loading"""
        self.assertGreater(len(self.meta.build_hook().global_action_list), 10)

    def test_multiple_edged_between_nodes_work(self):
        """Test that multiple edges between the same two nodes are allowed"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 3)
        edge_names = [edges[0].name, edges[1].name, edges[2].name]
        self.assertIn("generic", edge_names)
        self.assertIn("apple", edge_names)
        self.assertIn("banana", edge_names)

    def test_multiple_edged_between_nodes_delete_all(self):
        """Test that multiple edges deleted correctly if delete all"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey2)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 0)

    def test_multiple_edged_between_nodes_delete_all_specific(self):
        """Test that multiple edges deleted correctly if delete all"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey2b)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 1)

    def test_multiple_edged_between_nodes_delete_all_labeled(self):
        """Test that multiple edges deleted correctly if delete all"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey2c)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 3)

    def test_multiple_edged_between_nodes_delete_filtered(self):
        """Test that multiple edges deleted correctly if delete filtered"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey3)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 5)

    def test_generic_can_be_used_to_specify_generic_edges(self):
        """Test that generic edge tag works"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey4)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 2)

    def test_can_disconnect_multi_nodes_simultaneously(self):
        """Test disconnecting mutilpe nodes"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey5)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 2)

    def test_can_connect_multi_nodes_simultaneously(self):
        """Test connecting mutilpe nodes"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey6)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 4)

    def test_can_disconnect_multi_nodes_advanced(self):
        """Test disconnecting mutilpe nodes advanced"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edgey7)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(gph.get_all_edges())
        self.assertEqual(len(edges), 3)

    def test_accessing_edges_basic(self):
        """Test accessing Edges"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edge_access)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        edges = list(list(gph.get_all_edges()))
        if edges[0].name == "apple":
            self.assertEqual(edges[0].context["v1"], 7)
            self.assertEqual(edges[1].context["x1"], 8)
        else:
            self.assertEqual(edges[1].context["v1"], 7)
            self.assertEqual(edges[0].context["x1"], 8)

    def test_has_assign(self):
        """Test assignment on definition"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.has_assign)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        nodes = gph.get_all_nodes()
        self.assertEqual(len(nodes), 3)
        num = 0
        for i in nodes:
            if i.name == "testnode":
                self.assertEqual(i.context["a"], 8)
                num += 1
        self.assertEqual(num, 2)

    def test_global_get_set(self):
        """Test assignment on definition"""
        mast = self.meta.build_super_master()
        sent = Sentinel(m_id=mast.jid, h=mast._h)
        gph = Graph(m_id=mast.jid, h=mast._h)
        sent.register_code(jtc.set_get_global)
        test_walker = sent.run_architype("setter")
        test_walker.prime(gph)
        test_walker.run()
        test_walker2 = sent.run_architype("getter")
        test_walker2.prime(gph)
        test_walker2.run()
        self.assertEqual(test_walker2.context["a"], 59)

    def test_global_set_requires_admin(self):
        """Test assignment on definition"""
        mast = self.meta.build_master()
        sent = Sentinel(m_id=mast.jid, h=mast._h)
        gph = Graph(m_id=mast.jid, h=mast._h)
        sent.register_code(jtc.set_get_global2)
        test_walker = sent.run_architype("setter")
        test_walker.prime(gph)
        test_walker.run()
        test_walker2 = sent.run_architype("getter")
        test_walker2.prime(gph)
        test_walker2.run()
        self.assertEqual(test_walker2.context["a"], None)

    def test_sentinel_version_label(self):
        """Test sentinel version labeling"""
        mast = self.meta.build_master()
        sent = Sentinel(m_id=mast.jid, h=mast._h)
        sent.register_code(jtc.version_label)
        self.assertEqual(sent.version, "alpha-1.0")

    def test_visibility_builtins(self):
        """Test builtins to see into nodes and edges"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.visibility_builtins)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(2, len(test_walker.report[0][0].keys()))
        self.assertGreaterEqual(len(test_walker.report[0][1].keys()), 6)
        self.assertLess(len(test_walker.report[0][1].keys()), 14)
        self.assertGreaterEqual(len(test_walker.report[0][2].keys()), 13)

    def test_spawn_ctx_for_edges_nodes(self):
        """Test builtins to see into nodes and edges"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.spawn_ctx_edge_node)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0]["age"], 32)
        self.assertEqual(test_walker.report[1]["meeting_place"], "college")
        self.assertEqual(test_walker.report[2]["name"], "Jane")
        self.assertEqual(test_walker.report[3]["kind"], "sister")

    def test_filter_ctx_for_edges_nodes(self):
        """Test builtins to see into nodes and edges"""
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.filter_ctx_edge_node)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0]["age"], 30)
        self.assertEqual(len(test_walker.report[1]), 0)

    def test_null_handling(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.null_handleing)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0], True)
        self.assertEqual(test_walker.report[1], False)
        self.assertEqual(test_walker.report[2], True)
        self.assertEqual(test_walker.report[3], False)

    def test_bool_type_convert(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.bool_type_convert)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0], True)
        self.assertEqual(True, test_walker.report[1]["name"])

    def test_typecasts(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.typecasts)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual(test_walker.report[0], 7.6)
        self.assertEqual(test_walker.report[1], 7)
        self.assertEqual(test_walker.report[2], "7.6")
        self.assertEqual(test_walker.report[3], True)
        self.assertEqual(test_walker.report[4], 7.0)
        self.assertEqual(test_walker.report[5], "Types comes back correct")

    def test_typecast_error(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.typecasts_error)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertIn("Cannot get edges from 7.6", test_walker.runtime_errors[0])
        self.assertIn(
            "Invalid cast of JAC_TYPE.STR to JAC_TYPE.INT",
            test_walker.runtime_errors[1],
        )

    def test_filter_on_context(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.filter_on_context)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertEqual({"yo": "Yeah i said"}, test_walker.report[0][0])
        self.assertNotIn("name", test_walker.report[0][1].keys())
        self.assertIn("name", test_walker.report[0][2].keys())

    def test_string_manipulation(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.string_manipulation)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], "t")
        self.assertEqual(rep[1], "tin")
        self.assertEqual(rep[2], "sting me ")
        self.assertEqual(rep[3], " TESTING ME  ")
        self.assertEqual(rep[4], " testing me  ")
        self.assertEqual(rep[5], " Testing Me  ")
        self.assertEqual(rep[6], " testing me  ")
        self.assertEqual(rep[7], " TeSTING ME  ")
        self.assertEqual(rep[8], False)
        self.assertEqual(rep[9], False)
        self.assertEqual(rep[10], False)
        self.assertEqual(rep[11], False)
        self.assertEqual(rep[12], False)
        self.assertEqual(rep[13], False)
        self.assertEqual(rep[14], False)
        self.assertEqual(rep[15], {"a": 5})
        self.assertEqual(rep[16], 2)
        self.assertEqual(rep[17], 5)
        self.assertEqual(rep[18], ["tEsting", "me"])
        self.assertEqual(rep[19], [" t", "sting me  "])
        self.assertEqual(rep[20], False)
        self.assertEqual(rep[21], False)
        self.assertEqual(rep[22], " tEsting you  ")
        self.assertEqual(rep[23], "tEsting me")
        self.assertEqual(rep[24], "Esting me")
        self.assertEqual(rep[25], "tEsting me  ")
        self.assertEqual(rep[26], "sting me  ")
        self.assertEqual(rep[27], " tEsting me")
        self.assertEqual(rep[28], " tEsting m")
        self.assertEqual(rep[29], True)

    def test_list_manipulation(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.list_manipulation)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], [4])
        self.assertEqual(rep[1], [5])
        self.assertEqual(rep[2], [4, 5, 5])
        self.assertEqual(rep[3], [5, 5, 4])
        self.assertEqual(rep[4], [4, 5, 5])
        self.assertEqual(rep[5], 2)
        self.assertEqual(rep[6], [5, 5, 4, 2])
        self.assertEqual(rep[7], [5, "apple", 4, 2])
        self.assertEqual(rep[8], 1)
        self.assertEqual(rep[9], [5, "apple", 4])
        self.assertEqual(rep[10], [])

    def test_list_reversed(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.list_reversed)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], [7, 2, 4])
        self.assertEqual(rep[1], [4, 2, 7])

    def test_dict_manipulation(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.dict_manipulation)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], {"four": 4, "five": 5})
        self.assertEqual(rep[1], {"four": 5, "five": 5})
        self.assertEqual(rep[2], [["four", 4], ["five", 5]])
        self.assertEqual(rep[3], ["four", "five"])
        self.assertEqual(rep[4], {"four": 4})
        self.assertEqual(rep[5], [4])
        self.assertEqual(rep[6], {"four": 7})
        self.assertEqual(rep[7], {})

    def test_string_join(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.string_join)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], "test_me_now")

    def test_sub_list(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.sub_list)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0][0], 5)
        self.assertEqual(rep[0][1], 6)
        self.assertEqual(rep[0][2], 7)

    def test_destroy_and_misc(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.destroy_and_misc)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], "Josh")
        self.assertEqual(
            rep[1], {"name": None, "age": 32, "birthday": None, "profession": None}
        )
        self.assertEqual(
            rep[2], {"age": 32, "birthday": None, "name": "pete", "profession": None}
        )
        self.assertEqual(rep[3], [1, 3])
        self.assertEqual(rep[4], {"a": "b"})
        self.assertEqual(rep[5], [1, 2, 5, 6, 7, 8, 9])
        self.assertEqual(rep[6], [1, 2, 45, 33, 7, 8, 9])
        self.assertEqual(rep[7], None)
        self.assertEqual(
            rep[8], {"age": 32, "birthday": None, "name": "pete", "profession": None}
        )
        self.assertEqual(rep[9], True)

    def test_arbitrary_assign_on_element(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.arbitrary_assign_on_element)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(
            rep[0], {"name": None, "age": None, "birthday": None, "profession": None}
        )

    def test_try_else_stmts(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.try_else_stmts)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(
            rep[0],
            {
                "args": ("division by zero",),
                "col": 15,
                "line": 4,
                "mod": "basic",
                "msg": "division by zero",
                "type": "ZeroDivisionError",
                "name": "init",
                "rule": "expression",
            },
        )
        self.assertEqual(rep[1], "dont need err")
        self.assertEqual(rep[2], None)
        self.assertEqual(rep[3], 2)

    def test_node_edge_same_name(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.node_edge_same_name)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep[0], {"meeting_place": "college"})
        self.assertEqual(
            rep[1], {"age": 32, "birthday": None, "name": "Josh", "profession": None}
        )

    def test_testcases(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.testcases)
        sent.run_tests(silent=True)
        self.assertEqual(len(sent.testcases), 4)
        for i in sent.testcases:
            self.assertEqual(i["passed"], True)

    def test_testcase_asserts(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        sent.register_code(jtc.testcase_asserts)
        sent.run_tests(silent=True)
        self.assertEqual(len(sent.testcases), 3)
        self.assertEqual(sent.testcases[0]["passed"], True)
        self.assertEqual(sent.testcases[1]["passed"], False)
        self.assertEqual(sent.testcases[2]["passed"], False)

    def test_report_not_to_jacset(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.report_not_to_jacset)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertIn("context", rep[0][0].keys())
        self.assertIn("j_type", rep[0][0].keys())

    def test_walker_spawn_unwrap_check(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.walker_spawn_unwrap_check)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertTrue(rep[0].startswith("urn:uuid"))

    def test_std_get_report(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.std_get_report)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep, [3, 5, 6, 7, [3, 5, 6, 7], 8])

    def test_func_with_array_index(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.func_with_array_index)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        rep = test_walker.report
        self.assertEqual(rep, [3, 5, 3])

    def test_rt_error_test1(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.rt_error_test1)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        self.assertIn("List index out of range", test_walker.runtime_errors[0])
        self.assertIn(" line ", test_walker.runtime_errors[0])
        self.assertIn(" col ", test_walker.runtime_errors[0])

    def test_root_type_nodes(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.root_type_nodes)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, ["root", "root"])

    def test_invalid_key_error(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.invalid_key_error)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        errors = test_walker.runtime_errors
        self.assertGreater(len(errors), 0)

    def test_auto_cast(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.auto_cast)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, [True, True])

    def test_no_error_on_dict_key_assign(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.no_error_on_dict_key_assign)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, [{"b": 4}])
        self.assertEqual(len(test_walker.runtime_errors), 0)

    def test_report_status(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.report_status)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, ["hello"])
        self.assertEqual(test_walker.report_status, 302)

    def test_graph_in_graph(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.graph_in_graph)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(len(report), 3)

    def test_min_max_on_list(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.min_max_on_list)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, [531.1, 3, 5, 1])

    def test_edge_bug(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.edge_bug)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(len(report[0]), 4)

    def test_rand_choice(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.rand_choice)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertIn(report[1], report[0])

    def test_struct_types(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(jtc.struct_types)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report[0], [43, 33])
        self.assertEqual(report[1], [33, 43])



import uuid
from unittest import TestCase

import jaseci.tests.jac_test_code as jtc
from jaseci.actor.sentinel import Sentinel
from jaseci.element.element import Element
from jaseci.graph.graph import Graph
from jaseci.graph.node import Node
from jaseci.svc import MetaService
from jaseci.utils.utils import TestCaseHelper, get_all_subclasses
from jaseci.actor.architype import Architype


class ArchitypeTests(TestCaseHelper, TestCase):
    def setUp(self):
        super().setUp()
        self.meta = MetaService()

    def tearDown(self):
        super().tearDown()

    def test_object_creation_basic_no_side_creation(self):
        """ """
        mast = self.meta.build_master()
        num_objs = len(mast._h.mem.keys())
        node1 = Node(m_id=mast._m_id, h=mast._h)
        node2 = Node(m_id=mast._m_id, h=mast._h, parent=node1)
        num_new = len(mast._h.mem.keys())
        self.assertEqual(num_new, num_objs + 2)

        new_graph = Graph(m_id=mast._m_id, h=mast._h)
        mast.graph_ids.add_obj(new_graph)
        num_new = len(mast._h.mem.keys())
        self.assertEqual(num_new, num_objs + 3)

        new_graph.attach_outbound(node1)
        new_graph.attach_outbound(node2)
        num_new = len(mast._h.mem.keys())
        self.assertEqual(num_new, num_objs + 5)

    def test_edge_removal_updates_nodes_edgelist(self):
        """ """
        mast = self.meta.build_master()
        node1 = Node(m_id=mast._m_id, h=mast._h)
        node2 = Node(m_id=mast._m_id, h=mast._h)
        edge = node1.attach_outbound(node2)
        self.assertEqual(len(node1.edge_ids), 1)
        self.assertEqual(len(node2.edge_ids), 1)
        self.assertEqual(len(edge), 1)
        edge[0].destroy()
        self.assertEqual(len(node1.edge_ids), 0)
        self.assertEqual(len(node2.edge_ids), 0)

    def test_object_creation_by_sentinel_no_leaks(self):
        """
        Test that the destroy of sentinels clears owned objects
        """
        mast = self.meta.build_master()
        num_objs = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        self.assertEqual(num_objs, 2)
        new_graph = Graph(m_id=mast._m_id, h=mast._h)
        sent = Sentinel(m_id=mast._m_id, h=mast._h)
        code = jtc.prog1
        mast.sentinel_ids.add_obj(sent)
        mast.graph_ids.add_obj(new_graph)
        num_new = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        self.assertEqual(num_new, num_objs + 2)

        sent.register_code(code)
        num_objs = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        sent.register_code(code)
        new_num = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        self.assertEqual(num_objs, new_num)

    def test_json_blob_of_objects(self):
        """
        Test saving object to json and back to python dict
        """
        for i in get_all_subclasses(Element):
            kwargs = {"m_id": "anon", "h": self.meta.build_hook()}
            orig = i(**kwargs)
            blob1 = orig.json(detailed=True)
            new = i(**kwargs)
            self.assertNotEqual(orig.id, new.id)
            new.json_load(blob1)
            self.assertEqual(orig.id, new.id)
            self.assertTrue(orig.is_equivalent(new))

    def test_supermaster_can_touch_all_data(self):
        mh = self.meta.build_hook()
        mast = self.meta.build_master(h=mh)
        mast2 = self.meta.build_master(h=mh)
        node12 = Node(m_id=mast2._m_id, h=mast2._h)
        supmast = self.meta.build_super_master(h=mh)
        bad = mh.get_obj(mast._m_id, uuid.UUID(node12.jid))
        good = mh.get_obj(supmast._m_id, uuid.UUID(node12.jid))
        self.assertEqual(good, node12)
        self.assertNotEqual(bad, node12)
        self.assertIsNone(bad)

    def test_id_list_smart_name_error(self):
        self.logger_on()
        mast = self.meta.build_master()
        sent = Sentinel(m_id=mast._m_id, h=mast._h)
        self.assertIn(
            "arch_ids", sent.arch_ids.obj_for_id_not_exist_error(uuid.UUID(int=0).urn)
        )

    def test_dont_store_invalid_feilds_in_blob(self):
        self.logger_on()
        mast = self.meta.build_master()
        sent = Sentinel(m_id=mast._m_id, h=mast._h)
        sent.fake_data = 5
        stored = sent.jsci_payload()
        sent2 = Sentinel(m_id=mast._m_id, h=mast._h)
        sent2.json_load(stored)
        self.assertNotIn("fake_data", vars(sent2).keys())

    def test_sentinel_default_archs_dont_grow(self):
        mast = self.meta.build_master()
        sent = Sentinel(m_id=mast._m_id, h=mast._h)
        sent.register_code(text="node simple; walker init {}")
        before = sent._h.get_object_distribution()[Architype]
        stored = sent.jsci_payload()
        sent2 = Sentinel(m_id=mast._m_id, h=mast._h)
        sent2.json_load(stored)
        sent2 = Sentinel(m_id=mast._m_id, h=mast._h)
        sent2.json_load(stored)
        after = sent2._h.get_object_distribution()[Architype]
        self.assertEqual(before, after)

    def test_sentinel_default_archs_dont_grow_multi_compile(self):
        mast = self.meta.build_master()
        sent = Sentinel(m_id=mast._m_id, h=mast._h)
        sent.register_code(text="node simple; walker init {}")
        before = sent._h.get_object_distribution()[Architype]
        stored = sent.jsci_payload()
        sent2 = Sentinel(m_id=mast._m_id, h=mast._h)
        sent2.json_load(stored)
        sent2.register_code(text="node simple; walker init {}")
        before_id = sent2.arch_ids[0]
        sent2.register_code(text="node simple; walker init {}")
        sent2.register_code(text="node simple; walker init {}")
        after_id = sent2.arch_ids[0]
        after = sent2._h.get_object_distribution()[Architype]
        self.assertEqual(before, after)
        self.assertNotEqual(before_id, after_id)

    def test_id_list_heals(self):
        mast = self.meta.build_master()
        sent = Sentinel(m_id=mast._m_id, h=mast._h)
        sent.register_code(text="node simple; walker init {}")
        before = len(sent.arch_ids)
        sent._h.get_obj(mast._m_id, uuid.UUID(sent.arch_ids[1])).destroy()
        sent._h.get_obj(mast._m_id, uuid.UUID(sent.arch_ids[3])).destroy()
        sent.arch_ids.obj_list()
        after = len(sent.arch_ids)
        self.assertEqual(after, before - 2)



"""
Interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.utils.utils import is_jsonable, parse_str_token, uuid_re
from jaseci.element.element import Element
from jaseci.graph.node import Node
from jaseci.graph.edge import Edge
from jaseci.attr.action import Action
from jaseci.jac.jac_set import JacSet
from jaseci.jac.ir.jac_code import jac_ast_to_ir, jac_ir_to_ast
from jaseci.jac.machine.jac_scope import JacScope
from jaseci.jac.machine.machine_state import MachineState, TryException

from jaseci.jac.machine.jac_value import JacValue
from jaseci.jac.machine.jac_value import jac_elem_unwrap as jeu
from copy import copy, deepcopy


class Interp(MachineState):
    """Shared interpreter class across both sentinels and walkers"""

    def run_attr_stmt(self, jac_ast, obj):
        """
        attr_stmt: has_stmt | can_stmt;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "has_stmt":
            self.run_has_stmt(kid[0], obj)
        #  Can statements in architype handled in architype load

    def run_has_stmt(self, jac_ast, obj):
        """
        has_stmt: KW_HAS has_assign (COMMA has_assign)* SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "has_assign":
                self.run_has_assign(i, obj)

    def run_has_assign(self, jac_ast, obj):
        """
        has_assign: KW_PRIVATE? KW_ANCHOR? (NAME | NAME EQ expression);
        """
        kid = self.set_cur_ast(jac_ast)
        while kid[0].name in ["KW_PRIVATE", "KW_ANCHOR"]:
            kid = kid[1:]
        var_name = kid[0].token_text()
        var_val = None  # jac's null
        if len(kid) > 1:
            var_val = self.run_expression(kid[2]).value
        if isinstance(obj, dict):
            obj[var_name] = var_val
        # Runs only once for walkers
        elif var_name not in obj.context.keys() or obj.j_type != "walker":
            JacValue(self, ctx=obj, name=var_name, value=var_val).write(
                kid[0], force=True
            )

    def run_can_stmt(self, jac_ast, obj):
        """
        can_stmt:
            KW_CAN dotted_name (preset_in_out event_clause)? (
                COMMA dotted_name (preset_in_out event_clause)?
            )* SEMI
            | KW_CAN NAME event_clause? code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        kid = kid[1:]
        while True:
            action_type = "activity"
            access_list = None
            preset_in_out = None
            if kid[0].name == "NAME":
                action_name = kid[0].token_text()
            else:
                action_name = self.run_dotted_name(kid[0])
            kid = kid[1:]
            if len(kid) > 0 and kid[0].name == "preset_in_out":
                preset_in_out = jac_ast_to_ir(kid[0])
                kid = kid[1:]
            if len(kid) > 0 and kid[0].name == "event_clause":
                action_type, access_list = self.run_event_clause(kid[0])
                kid = kid[1:]
            # if (not isinstance(obj, node) and action_type != 'activity'):
            #     self.rt_warn(
            #         "Only nodes has on entry/exit, treating as activity",
            #         kid[0])
            #     action_type = 'activity'
            if kid[0].name == "code_block":
                act = Action(
                    m_id=self._m_id,
                    h=self._h,
                    name=action_name,
                    value=jac_ast_to_ir(kid[0]),
                    preset_in_out=preset_in_out,
                    access_list=access_list,
                )
                getattr(obj, f"{action_type}_action_ids").add_obj(act)
                self._jac_scope.add_action(act)
                break
            else:
                # self.check_builtin_action(action_name, jac_ast)
                act = Action(
                    m_id=self._m_id,
                    h=self._h,
                    name=action_name,
                    value=action_name,
                    preset_in_out=preset_in_out,
                    access_list=access_list,
                )
                getattr(obj, f"{action_type}_action_ids").add_obj(act)
                self._jac_scope.add_action(act)
            if not len(kid) or kid[0].name != "COMMA":
                break
            else:
                kid = kid[1:]

    def run_event_clause(self, jac_ast):
        """
        event_clause:
                KW_WITH name_list? (KW_ENTRY | KW_EXIT | KW_ACTIVITY);
        """
        kid = self.set_cur_ast(jac_ast)
        nl = []
        if kid[1].name == "name_list":
            nl = self.run_name_list(kid[1])
        return kid[-1].token_text(), nl

    def run_dotted_name(self, jac_ast):
        """
        dotted_name: NAME (DOT NAME)*;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = ""
        for i in kid:
            if i.name == "NAME":
                ret += i.token_text()
                if i == kid[-1]:
                    break
                ret += "."
        return ret

    def run_name_list(self, jac_ast):
        """
        name_list: NAME (COMMA NAME)*;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = []
        for i in kid:
            if i.name == "NAME":
                ret.append(i.token_text())
        return ret

    def run_expr_list(self, jac_ast, wrap=False):
        """
        expr_list: expression (COMMA expression)*;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = []
        for i in kid:
            if i.name == "expression":
                if wrap:
                    ret.append(self.run_expression(i).wrap())
                else:
                    ret.append(self.run_expression(i).value)
        return JacValue(self, value=ret)

    def run_code_block(self, jac_ast):
        """
        code_block: LBRACE statement* RBRACE | COLON statement;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "statement" and not self._loop_ctrl:
                self.run_statement(jac_ast=i)

    def run_statement(self, jac_ast):
        """
        statement:
            code_block
            | node_ctx_block
            | expression SEMI
            | if_stmt
            | try_stmt
            | for_stmt
            | while_stmt
            | assert_stmt SEMI
            | ctrl_stmt SEMI
            | destroy_action
            | report_action
            | walker_action;
        """
        if self._stopped:
            return
        kid = self.set_cur_ast(jac_ast)
        self.run_rule(kid[0])

    def run_if_stmt(self, jac_ast):
        """
        if_stmt: KW_IF expression code_block elif_stmt* else_stmt?;
        """
        kid = self.set_cur_ast(jac_ast)
        if self.run_expression(kid[1]).value:
            self.run_code_block(kid[2])
            return
        kid = kid[3:]
        if len(kid):
            while True:
                if kid[0].name == "elif_stmt":
                    if self.run_elif_stmt(kid[0]):
                        return
                elif kid[0].name == "else_stmt":
                    self.run_else_stmt(kid[0])
                    return
                kid = kid[1:]
                if not len(kid):
                    break

    def run_try_stmt(self, jac_ast):
        """
        try_stmt: KW_TRY code_block else_from_try?;
        """
        kid = self.set_cur_ast(jac_ast)
        try:
            self.run_code_block(kid[1])
            return
        except TryException as e:
            if len(kid) > 2:
                self.run_else_from_try(kid[2], e.ref)
        except Exception as e:
            if len(kid) > 2:
                self.run_else_from_try(kid[2], self.jac_exception(e, kid[2]))

    def run_else_from_try(self, jac_ast, jac_ex):
        """
        else_from_try:
            KW_ELSE (LPAREN NAME RPAREN)? code_block
            | KW_ELSE (KW_WITH NAME)? code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        if len(kid) > 2:
            JacValue(
                self,
                ctx=self._jac_scope.local_scope,
                name=kid[2].token_text(),
                value=jac_ex,
            ).write(kid[2])
        self.run_code_block(kid[-1])

    def run_elif_stmt(self, jac_ast):
        """
        elif_stmt: KW_ELIF expression code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        if self.run_expression(kid[1]).value:
            self.run_code_block(kid[2])
            return True
        else:
            return False

    def run_else_stmt(self, jac_ast):
        """
        else_stmt: KW_ELSE code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        self.run_code_block(kid[1])

    def run_for_stmt(self, jac_ast):
        """
        for_stmt:
            KW_FOR expression KW_TO expression KW_BY expression code_block
            | KW_FOR NAME (COMMA NAME)? KW_IN expression code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        loops = 0
        if kid[1].name == "expression":
            self.run_expression(kid[1])
            while self.run_expression(kid[3]).value:
                self._loop_ctrl = None
                self.run_code_block(kid[6])
                self.run_expression(kid[5])
                loops += 1
                if self._loop_ctrl and self._loop_ctrl == "break":
                    self._loop_ctrl = None
                    break
                if loops > self._loop_limit:
                    self.rt_error("Hit loop limit, breaking...", kid[0])
                    self._loop_ctrl = "break"
        elif kid[3].name == "expression":
            var = self._jac_scope.get_live_var(kid[1].token_text(), create_mode=True)
            lst = self.run_expression(kid[3]).value

            if isinstance(lst, (list, dict)):
                for i in lst:
                    self._loop_ctrl = None
                    var.value = i
                    var.write(kid[1])
                    self.run_code_block(kid[4])
                    loops += 1
                    if self._loop_ctrl and self._loop_ctrl == "break":
                        self._loop_ctrl = None
                        break
                    if loops > self._loop_limit:
                        self.rt_error("Hit loop limit, breaking...", kid[0])
                        self._loop_ctrl = "break"
            else:
                self.rt_error("Not a list/dict for iteration!", kid[3])
        else:
            key = self._jac_scope.get_live_var(kid[1].token_text(), create_mode=True)
            val = self._jac_scope.get_live_var(kid[3].token_text(), create_mode=True)
            source = self.run_expression(kid[5]).value

            lst = None
            if isinstance(source, dict):
                lst = source.items()
            elif isinstance(source, list):
                lst = enumerate(source)

            if not (lst is None):
                for k, v in lst:
                    self._loop_ctrl = None
                    key.value = k
                    key.write(kid[1])
                    val.value = v
                    val.write(kid[3])
                    self.run_code_block(kid[6])
                    loops += 1
                    if self._loop_ctrl and self._loop_ctrl == "break":
                        self._loop_ctrl = None
                        break
                    if loops > self._loop_limit:
                        self.rt_error("Hit loop limit, breaking...", kid[0])
                        self._loop_ctrl = "break"
            else:
                self.rt_error("Not a list/dict for iteration!", kid[5])
        if self._loop_ctrl and self._loop_ctrl == "continue":
            self._loop_ctrl = None

    def run_while_stmt(self, jac_ast):
        """
        while_stmt: KW_WHILE expression code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        loops = 0
        while self.run_expression(kid[1]).value:
            self._loop_ctrl = None
            self.run_code_block(kid[2])
            loops += 1
            if self._loop_ctrl and self._loop_ctrl == "break":
                self._loop_ctrl = None
                break
            if loops > self._loop_limit:
                self.rt_error("Hit loop limit, breaking...", kid[0])
                self._loop_ctrl = "break"
        if self._loop_ctrl and self._loop_ctrl == "continue":
            self._loop_ctrl = None

    def run_ctrl_stmt(self, jac_ast):
        """
        ctrl_stmt: KW_CONTINUE | KW_BREAK | KW_SKIP;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "KW_SKIP":
            self._stopped = "skip"
        elif kid[0].name == "KW_BREAK":
            self._loop_ctrl = "break"
        elif kid[0].name == "KW_CONTINUE":
            self._loop_ctrl = "continue"

    def run_assert_stmt(self, jac_ast):
        """
        assert_stmt: KW_ASSERT expression;
        """
        kid = self.set_cur_ast(jac_ast)
        passed = False
        try:
            passed = self.run_expression(kid[1]).value
        except Exception as e:
            e = self.jac_exception(e, jac_ast)
            raise Exception("Jac Assert Failed", kid[1].get_text(), e)
        if not passed:
            raise Exception("Jac Assert Failed", kid[1].get_text())

    def run_destroy_action(self, jac_ast):
        """
        destroy_action: KW_DESTROY expression SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1])
        if isinstance(result.value, Element):
            self.destroy_node_ids.add_obj(result.value)
        elif isinstance(result.value, JacSet):
            self.destroy_node_ids.add_obj_list(result.value)
        result.self_destruct(kid[1])

    def run_report_action(self, jac_ast):
        """
        report_action:
            KW_REPORT expression SEMI
                | KW_REPORT COLON NAME EQ expression SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[1].name == "COLON":
            if kid[2].token_text() in ["status", "status_code"]:
                self.report_status = self.run_expression(kid[4]).value
            elif kid[2].token_text() == "custom":
                self.report_custom = self.run_expression(kid[4]).value
            elif kid[2].token_text() == "error":
                err = self.run_expression(kid[4]).value
                if isinstance(err, str):
                    self.runtime_errors.append(err)
                else:
                    self.runtime_errors.append(
                        f'{err["mod"]}:{err["name"]} - line {err["line"]}, '
                        + f'col {err["col"]} - rule {err["rule"]} - {err["msg"]}'
                    )
            else:
                self.rt_error("Invalid report attribute to set", kid[2])
        else:
            report = self.run_expression(kid[1]).wrap(serialize_mode=True)
            if not is_jsonable(report):
                self.rt_error("Report not Json serializable", kid[0])
            self.report.append(copy(report))

    def run_expression(self, jac_ast):
        """
        expression: connect (assignment | copy_assign | inc_assign)?;
        """
        try:

            def check_can_write(val):
                if val.ctx is None:
                    self.rt_error("Cannot assign to this expression", kid[0])
                    return False
                return True

            kid = self.set_cur_ast(jac_ast)

            if len(kid) == 1:
                return self.run_rule(kid[0])
            else:
                if kid[1].name == "assignment":
                    self._assign_mode = True
                    dest = self.run_rule(kid[0])
                    self._assign_mode = False
                    if not check_can_write(dest):
                        return dest
                    return self.run_assignment(kid[1], dest=dest)
                elif kid[1].name == "copy_assign":
                    self._assign_mode = True
                    dest = self.run_rule(kid[0])
                    self._assign_mode = False
                    if not check_can_write(dest):
                        return dest
                    return self.run_copy_assign(kid[1], dest=dest)
                elif kid[1].name == "inc_assign":
                    self._assign_mode = True
                    dest = self.run_rule(kid[0])
                    self._assign_mode = False
                    if not check_can_write(dest):
                        return dest
                    return self.run_inc_assign(kid[1], dest=dest)
        except Exception as e:
            self.jac_try_exception(e, jac_ast)

    def run_assignment(self, jac_ast, dest):
        """
        assignment: EQ expression;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1])
        dest.value = result.value
        dest.write(jac_ast)
        return dest

    def run_copy_assign(self, jac_ast, dest):
        """
        copy_assign: CPY_EQ expression;
        """
        kid = self.set_cur_ast(jac_ast)
        src = self.run_expression(kid[1])
        if not self.rt_check_type(dest.value, [Node, Edge], kid[1]):
            self.rt_error("':=' only applies to nodes and edges", kid[1])
            return dest
        if dest.value.name != src.value.name:
            self.rt_error(
                f"Node/edge arch {dest.value} don't " f"match {src.value}!", kid[1]
            )
            return dest
        for i in src.value.context.keys():
            if i in dest.value.context.keys():
                JacValue(
                    self, ctx=dest.value, name=i, value=src.value.context[i]
                ).write(jac_ast)
        return dest

    def run_inc_assign(self, jac_ast, dest):
        """
        inc_assign: (PEQ | MEQ | TEQ | DEQ) expression;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "PEQ":
            dest.value = dest.value + self.run_expression(kid[1]).value
        elif kid[0].name == "MEQ":
            dest.value = dest.value - self.run_expression(kid[1]).value
        elif kid[0].name == "TEQ":
            dest.value = dest.value * self.run_expression(kid[1]).value
        elif kid[0].name == "DEQ":
            dest.value = dest.value / self.run_expression(kid[1]).value
        dest.write(jac_ast)
        return dest

    def run_connect(self, jac_ast):
        """
        connect: logical ( (NOT)? edge_ref expression)?;
        """
        kid = self.set_cur_ast(jac_ast)
        if len(kid) < 2:
            return self.run_rule(kid[0])
        bret = self.run_rule(kid[0])
        base = bret.value
        tret = self.run_expression(kid[-1])
        target = tret.value
        self.rt_check_type(base, [Node, JacSet], kid[0])
        self.rt_check_type(target, [Node, JacSet], kid[-1])
        if isinstance(base, Node):
            base = JacSet(in_list=[base])
        if isinstance(target, Node):
            target = JacSet(in_list=[target])
        if kid[1].name == "NOT":
            for i in target.obj_list():
                for j in base.obj_list():
                    j.detach_edges(i, self.run_edge_ref(kid[2]).obj_list())
            return bret
        else:
            direction = kid[1].kid[0].name
            for i in target.obj_list():
                for j in base.obj_list():
                    use_edge = self.run_edge_ref(kid[1], is_spawn=True)
                    self.rt_check_type(i, Node, kid[-1])
                    self.rt_check_type(j, Node, kid[-1])
                    if direction == "edge_from":
                        j.attach_inbound(i, [use_edge])
                    elif direction == "edge_to":
                        j.attach_outbound(i, [use_edge])
                    else:
                        j.attach_bidirected(i, [use_edge])
        return bret

    def run_logical(self, jac_ast):
        """
        logical: compare ((KW_AND | KW_OR) compare)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_rule(kid[0])
        kid = kid[1:]
        while kid:
            if kid[0].name == "KW_AND":
                if result.value:
                    result.value = result.value and self.run_rule(kid[1]).value
            elif kid[0].name == "KW_OR":
                if not result.value:
                    result.value = result.value or self.run_rule(kid[1]).value
            kid = kid[2:]
            if not kid:
                break
        return result

    def run_compare(self, jac_ast):
        """
        compare: NOT compare | arithmetic (cmp_op arithmetic)*;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "NOT":
            return JacValue(self, value=not self.run_rule(kid[1]).value)
        else:
            result = self.run_rule(kid[0])
            kid = kid[1:]
            while kid:
                other_res = self.run_rule(kid[1])
                result = self.run_cmp_op(kid[0], result, other_res)
                kid = kid[2:]
                if not kid:
                    break
            return result

    def run_cmp_op(self, jac_ast, val1, val2):
        """
        cmp_op: EE | LT | GT | LTE | GTE | NE | KW_IN | nin;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "EE":
            return JacValue(self, value=val1.value == val2.value)
        elif kid[0].name == "LT":
            return JacValue(self, value=val1.value < val2.value)
        elif kid[0].name == "GT":
            return JacValue(self, value=val1.value > val2.value)
        elif kid[0].name == "LTE":
            return JacValue(self, value=val1.value <= val2.value)
        elif kid[0].name == "GTE":
            return JacValue(self, value=val1.value >= val2.value)
        elif kid[0].name == "NE":
            return JacValue(self, value=val1.value != val2.value)
        elif kid[0].name == "KW_IN":
            return JacValue(self, value=val1.value in val2.value)
        elif kid[0].name == "nin":
            return JacValue(self, value=val1.value not in val2.value)

    def run_arithmetic(self, jac_ast):
        """
        arithmetic: term ((PLUS | MINUS) term)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_rule(kid[0])
        kid = kid[1:]
        while kid:
            other_res = self.run_rule(kid[1])
            if kid[0].name == "PLUS":
                result.value = result.value + other_res.value
            elif kid[0].name == "MINUS":
                result.value = result.value - other_res.value
            kid = kid[2:]
            if not kid:
                break
        return result

    def run_term(self, jac_ast):
        """
        term: factor ((STAR_MUL | DIV | MOD) factor)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_rule(kid[0])
        kid = kid[1:]
        while kid:
            other_res = self.run_rule(kid[1])
            if kid[0].name == "STAR_MUL":
                result.value = result.value * other_res.value
            elif kid[0].name == "DIV":
                result.value = result.value / other_res.value
            elif kid[0].name == "MOD":
                result.value = result.value % other_res.value
            kid = kid[2:]
            if not kid:
                break
        return result

    def run_factor(self, jac_ast):
        """
        factor: (PLUS | MINUS) factor | power;
        """
        kid = self.set_cur_ast(jac_ast)
        if len(kid) < 2:
            return self.run_rule(kid[0])
        else:
            result = self.run_rule(kid[1])
            if kid[0].name == "MINUS":
                result.value = -(result.value)
            return result

    def run_power(self, jac_ast):
        """
        power: atom (POW factor)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_rule(kid[0])
        kid = kid[1:]
        if len(kid) < 1:
            return result
        elif kid[0].name == "POW":
            while kid:
                result.value = result.value ** self.run_rule(kid[1]).value
                kid = kid[2:]
                if not kid:
                    break
            return result

    def run_global_ref(self, jac_ast):
        """
        global_ref: KW_GLOBAL DOT (obj_built_in | NAME);
        """
        try:
            kid = self.set_cur_ast(jac_ast)
            if kid[2].name == "obj_built_in":
                kid = self.set_cur_ast(kid[2])
                if kid[0].name == "KW_CONTEXT":
                    return JacValue(self, value=self.parent().global_vars)
                elif kid[0].name == "KW_INFO":
                    # Add additional accessible fields
                    return JacValue(self, value=self.get_info())
                else:
                    self.rt_error(f"Global {kid[0].name} not yet", jac_ast)
                    return JacValue(
                        self,
                    )
            else:
                token = kid[2].token_text()
                if token not in self.parent().global_vars:
                    self.rt_error(f"Global not defined - {token}", kid[2])
                    return JacValue(
                        self,
                    )

                return JacValue(self, ctx=self.parent().global_vars, name=token)

        except Exception as e:
            self.jac_try_exception(e, jac_ast)

    def run_atom(self, jac_ast):
        """
        atom:
            INT
            | FLOAT
            | STRING
            | BOOL
            | NULL
            | NAME
            | global_ref
            | node_edge_ref
            | type_ref spawn_ctx?
            | list_val
            | dict_val
            | LPAREN expression RPAREN
            | ability_op NAME spawn_ctx?
            | atom atom_trailer+
            | KW_SYNC atom
            | spawn
            | ref
            | deref
            | any_type;
        """
        try:
            kid = self.set_cur_ast(jac_ast)
            if kid[0].name == "INT":
                return JacValue(self, value=int(kid[0].token_text()))
            elif kid[0].name == "FLOAT":
                return JacValue(self, value=float(kid[0].token_text()))
            elif kid[0].name == "STRING":
                return JacValue(self, value=parse_str_token(kid[0].token_text()))
            elif kid[0].name == "BOOL":
                return JacValue(self, value=bool(kid[0].token_text() == "true"))
            elif kid[0].name == "NULL":
                return JacValue(self, value=None)
            elif kid[0].name == "NAME":
                name = kid[0].token_text()
                val = self._jac_scope.get_live_var(name, create_mode=self._assign_mode)
                if val is None:
                    self.rt_error(f"Variable not defined - {name}", kid[0])
                    return JacValue(self)
                return val
            elif kid[0].name == "type_ref":
                ret = self.run_type_ref(kid[0])
                if kid[-1].name == "spawn_ctx":
                    self.run_spawn_ctx(kid[-1], ret.value)
                return ret
            elif kid[0].name == "LPAREN":
                return self.run_expression(kid[1])
            elif kid[0].name == "ability_op":
                return self.run_atom_trailer(jac_ast, None)
            elif kid[0].name == "atom":
                ret = self.run_atom(kid[0])
                for i in kid[1:]:
                    ret = self.run_atom_trailer(i, ret)
                return ret
            elif kid[0].name == "KW_SYNC":
                val = self.run_atom(kid[1])
                task_func = self._h.task
                if not task_func.is_running():
                    raise Exception("Task hook is not yet initialized!")
                return JacValue(
                    self, value=task_func.get_by_task_id(val.value["result"], True)
                )
            else:
                return self.run_rule(kid[0])

        except Exception as e:
            self.jac_try_exception(e, jac_ast)

    def run_atom_trailer(self, jac_ast, atom_res):
        """
        atom_trailer:
            DOT built_in
            | DOT NAME
            | index_slice
            | LPAREN expr_list? RPAREN
            | ability_op NAME spawn_ctx?;
        """
        try:
            kid = self.set_cur_ast(jac_ast)
            if atom_res is None:
                atom_res = JacValue(self, value=self._jac_scope.has_obj)
            if kid[0].name == "DOT":
                if kid[1].name == "built_in":

                    return self.run_built_in(kid[1], atom_res)
                elif kid[1].name == "NAME":
                    d = atom_res.value
                    n = kid[1].token_text()
                    if self.rt_check_type(d, [dict, Element, JacSet], kid[0]):
                        if not isinstance(d, JacSet):
                            ret = JacValue(self, ctx=d, name=n)
                        else:
                            plucked = []
                            for i in d:
                                if n in i.context.keys():
                                    plucked.append(i.context[n])
                                else:
                                    self.rt_error(
                                        f"Some elements in set does not have {n}",
                                        kid[1],
                                    )
                            ret = JacValue(self, value=plucked)
                        ret.unwrap()
                        return ret
                    else:
                        self.rt_error(f"Invalid variable {n}", kid[0])
            elif kid[0].name == "index_slice":
                if not self.rt_check_type(atom_res.value, [list, str, dict], kid[0]):
                    return atom_res
                return self.run_index_slice(kid[0], atom_res)
            elif kid[0].name == "LPAREN":
                param_list = []
                if kid[1].name == "expr_list":
                    param_list = self.run_expr_list(kid[1]).value
                if isinstance(atom_res.value, Action):
                    ret = atom_res.value.trigger(param_list, self._jac_scope, self)
                    return JacValue(self, value=ret)
                else:
                    self.rt_error("Unable to execute ability", kid[0])
            elif kid[0].name == "ability_op":
                arch = self.run_ability_op(kid[0], atom_res)
                if len(kid) > 2:
                    self.run_spawn_ctx(kid[2], atom_res.value)
                self.call_ability(
                    nd=atom_res.value,
                    name=kid[1].token_text(),
                    act_list=arch.get_all_actions(),
                )
                return atom_res
        except Exception as e:
            self.jac_try_exception(e, jac_ast)

    def run_ability_op(self, jac_ast, atom_res):
        """
        ability_op: DBL_COLON | DBL_COLON NAME COLON;
        """
        kid = self.set_cur_ast(jac_ast)
        base_arch = atom_res.value.get_architype()
        if len(kid) > 1:
            kind = atom_res.value.kind
            name = kid[1].token_text()
            if name in base_arch.derived_types():
                return self.parent().arch_ids.get_obj_by_name(name=name, kind=kind)
            else:
                self.rt_error(f"{name} is not a super arch of {base_arch.name}")
                return None
        else:
            return base_arch

    def run_ref(self, jac_ast):
        """
        ref: '&' atom;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_atom(kid[1])
        if self.rt_check_type(result.value, Element, kid[1]):
            result = JacValue(self, value=result.value.jid)
        return result

    def run_deref(self, jac_ast):
        """
        deref: '*' atom;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_atom(kid[1])

        if (
            isinstance(result.value, str)
            and len(result.value) < 64  # super long string, untrustworthy
            and not result.value.startswith("jac:uuid:")  # already an object
        ):
            matcher = uuid_re.search(result.value)
            if matcher and matcher.group(1):
                nd = jeu(f"jac:uuid:{matcher.group(1)}", self)
                if nd is not None:
                    return JacValue(self, value=nd)

        self.rt_error(f"{result.value} not valid reference", kid[1])
        return result

    def run_built_in(self, jac_ast, atom_res):
        """
        built_in:
            cast_built_in
            | obj_built_in
            | dict_built_in
            | list_built_in
            | string_built_in;
        """
        return self.run_rule(jac_ast.kid[0], atom_res)

    def run_cast_built_in(self, jac_ast, atom_res):
        """
        arch_built_in: any_type;
        """
        kid = self.set_cur_ast(jac_ast)
        typ = self.run_any_type(kid[0])
        if typ.value == Edge:
            if isinstance(atom_res.value, Node):
                return JacValue(
                    self,
                    value=self.obj_set_to_jac_set(
                        self.current_node.attached_edges(atom_res.value)
                    ),
                )
            elif isinstance(atom_res.value, Edge):
                return atom_res
            elif isinstance(atom_res.value, JacSet):
                return JacValue(self, value=self._relevant_edges)
                # res = jac_set()
                # for i in atom_res.value.obj_list():
                #     if(isinstance(i, edge)):
                #         res.add_obj(i)
                #     elif(isinstance(i, node)):
                #         res += self.obj_set_to_jac_set(
                #             self.current_node.attached_edges(i))
                # return jac_value(self, value=res)
            else:
                self.rt_error(
                    f"Cannot get edges from {atom_res.value}. "
                    f"Type {atom_res.jac_type()} invalid",
                    kid[0],
                )
        # may want to remove 'here" node from return below
        elif typ.value == Node:
            if isinstance(atom_res.value, Node):
                return atom_res
            elif isinstance(atom_res.value, Edge):
                return JacValue(self, value=self.obj_set_to_jac_set(atom_res.nodes()))
            elif isinstance(atom_res.value, JacSet):
                res = JacSet()
                for i in atom_res.value.obj_list():
                    if isinstance(i, Edge):
                        res.add_obj(i.to_node())
                        res.add_obj(i.from_node())
                    elif isinstance(i, Node):
                        res.add_obj(i)
                return JacValue(self, value=res)
            else:
                self.rt_error(
                    f"Cannot get nodes from {atom_res}. "
                    f"Type {atom_res.jac_type()} invalid",
                    kid[0],
                )
        else:
            try:
                atom_res.value = typ.value(atom_res.value)
            except Exception:
                self.rt_error(
                    f"Invalid cast of {atom_res.jac_type()} " f"to {typ.wrap()}", kid[0]
                )
            return atom_res

        return atom_res

    def run_obj_built_in(self, jac_ast, atom_res):
        """
        obj_built_in: KW_CONTEXT | KW_INFO | KW_DETAILS;
        """
        kid = self.set_cur_ast(jac_ast)
        from jaseci.actor.walker import Walker

        if kid[0].name == "KW_CONTEXT":
            if self.rt_check_type(atom_res.value, [Node, Edge, Walker], kid[0]):
                return JacValue(self, value=atom_res.value.context)
        elif kid[0].name == "KW_INFO":
            if self.rt_check_type(atom_res.value, [Node, Edge, Walker], kid[0]):
                return JacValue(self, value=atom_res.value.serialize(detailed=False))
        elif kid[0].name == "KW_DETAILS":
            if self.rt_check_type(atom_res.value, [Node, Edge, Walker], kid[0]):
                return JacValue(self, value=atom_res.value.serialize(detailed=True))
        return atom_res

    def run_dict_built_in(self, jac_ast, atom_res):
        """
        dict_built_in:
            KW_KEYS
            | LBRACE name_list RBRACE
            | (TYP_DICT DBL_COLON | DICT_DBL_COLON) NAME (
                LPAREN expr_list RPAREN
            )?;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "KW_KEYS":
            if isinstance(atom_res.value, dict):
                return JacValue(self, value=list(atom_res.value.keys()))
            else:
                self.rt_error(
                    f"Cannot get keys of {atom_res}. " f"Not Dictionary!", kid[0]
                )
                return JacValue(self, value=[])
        elif len(kid) > 1 and kid[1].name == "name_list":
            filter_on = self.run_name_list(kid[1])
            d = atom_res.value
            if self.rt_check_type(d, [dict], kid[0]):
                d = {k: d[k] for k in d if k in filter_on}
                return JacValue(self, value=d)
        else:
            if not self.rt_check_type(atom_res.value, [dict], kid[0]):
                return atom_res
            kid = kid[1:]
            if kid[0].name == "DBL_COLON":
                kid = kid[1:]
            result = None
            op = kid[0].token_text()
            try:
                if op == "items":
                    result = JacValue(
                        self, value=list(map(list, atom_res.value.items()))
                    )
                elif op == "copy":
                    result = JacValue(self, value=atom_res.value.copy())
                elif op == "deepcopy":
                    result = JacValue(self, value=deepcopy(atom_res.value))
                elif op == "keys":
                    result = JacValue(self, value=list(atom_res.value.keys()))
                elif op == "clear":
                    result = JacValue(self, value=atom_res.value.clear())
                elif op == "popitem":
                    result = JacValue(self, value=list(atom_res.value.popitem()))
                elif op == "values":
                    result = JacValue(self, value=list(atom_res.value.values()))
                if result:
                    if len(kid) > 1:
                        self.rt_warn(f"{op} does not take parameters, ignoring", kid[2])
                    return result
                if len(kid) > 1:
                    args = self.run_expr_list(kid[2]).value
                    if op == "pop":
                        result = JacValue(self, value=atom_res.value.pop(*args))
                    elif op == "update":
                        result = JacValue(self, value=atom_res.value.update(*args))
                    elif op == "get":
                        result = JacValue(self, value=atom_res.value.get(*args))
                    if result:
                        return result
            except Exception as e:
                self.rt_error(f"{e}", jac_ast)
            self.rt_error(f"Call to {op} is invalid.", jac_ast)

        return atom_res

    def run_list_built_in(self, jac_ast, atom_res):
        """
        list_built_in:
            KW_LENGTH
            | (TYP_LIST DBL_COLON | LIST_DBL_COLON) NAME (
                LPAREN expr_list RPAREN
            )?;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "KW_LENGTH":
            if isinstance(atom_res.value, list):
                return JacValue(self, value=len(atom_res.value))
            else:
                self.rt_error(
                    f"Cannot get length of {atom_res.value}. Not List!", kid[0]
                )
                return JacValue(self, value=0)
        else:
            if not self.rt_check_type(atom_res.value, [list], kid[0]):
                return atom_res
            kid = kid[1:]
            if kid[0].name == "DBL_COLON":
                kid = kid[1:]
            result = None
            op = kid[0].token_text()
            try:
                if op == "reverse":
                    result = JacValue(self, value=atom_res.value.reverse())
                elif op == "reversed":
                    result = JacValue(self, value=list(reversed(atom_res.value)))
                elif op == "copy":
                    result = JacValue(self, value=atom_res.value.copy())
                elif op == "deepcopy":
                    result = JacValue(self, value=deepcopy(atom_res.value))
                elif op == "sort":
                    result = JacValue(self, value=atom_res.value.sort())
                elif op == "clear":
                    result = JacValue(self, value=atom_res.value.clear())
                elif op == "max":
                    result = JacValue(self, value=max(atom_res.value))
                elif op == "min":
                    result = JacValue(self, value=min(atom_res.value))
                elif op == "idx_of_max":
                    result = JacValue(
                        self, value=atom_res.value.index(max(atom_res.value))
                    )
                elif op == "idx_of_min":
                    result = JacValue(
                        self, value=atom_res.value.index(min(atom_res.value))
                    )
                elif len(kid) < 2 and op == "pop":
                    result = JacValue(self, value=atom_res.value.pop())
                if result:
                    if len(kid) > 1:
                        self.rt_warn(f"{op} does not take parameters, ignoring", kid[2])
                    return result
                if len(kid) > 1:
                    args = self.run_expr_list(kid[2]).value
                    if op == "index":
                        result = JacValue(self, value=atom_res.value.index(*args))
                    elif op == "append":
                        result = JacValue(self, value=atom_res.value.append(*args))
                    elif op == "extend":
                        result = JacValue(self, value=atom_res.value.extend(*args))
                    elif op == "insert":
                        result = JacValue(self, value=atom_res.value.insert(*args))
                    elif op == "remove":
                        result = JacValue(self, value=atom_res.value.remove(*args))
                    elif op == "count":
                        result = JacValue(self, value=atom_res.value.count(*args))
                    elif op == "pop":
                        result = JacValue(self, value=atom_res.value.pop(*args))
                    if result:
                        return result
            except Exception as e:
                self.rt_error(f"{e}", jac_ast)
            self.rt_error(f"Call to {op} is invalid.", jac_ast)
        return atom_res

    def run_string_built_in(self, jac_ast, atom_res):
        """
        string_built_in:
        (TYP_STRING DBL_COLON | STR_DBL_COLON) NAME (
                LPAREN expr_list RPAREN
        )?;
        """
        kid = self.set_cur_ast(jac_ast)
        if not self.rt_check_type(atom_res.value, [str], kid[0]):
            return atom_res
        kid = kid[1:]
        if kid[0].name == "DBL_COLON":
            kid = kid[1:]
        result = None
        str_op = kid[0].token_text()
        try:
            if str_op == "upper":
                result = JacValue(self, value=atom_res.value.upper())
            elif str_op == "lower":
                result = JacValue(self, value=atom_res.value.lower())
            elif str_op == "title":
                result = JacValue(self, value=atom_res.value.title())
            elif str_op == "capitalize":
                result = JacValue(self, value=atom_res.value.capitalize())
            elif str_op == "swap_case":
                result = JacValue(self, value=atom_res.value.swapcase())
            elif str_op == "is_alnum":
                result = JacValue(self, value=atom_res.value.isalnum())
            elif str_op == "is_alpha":
                result = JacValue(self, value=atom_res.value.isalpha())
            elif str_op == "is_digit":
                result = JacValue(self, value=atom_res.value.isdigit())
            elif str_op == "is_title":
                result = JacValue(self, value=atom_res.value.istitle())
            elif str_op == "is_upper":
                result = JacValue(self, value=atom_res.value.isupper())
            elif str_op == "is_lower":
                result = JacValue(self, value=atom_res.value.islower())
            elif str_op == "is_space":
                result = JacValue(self, value=atom_res.value.isspace())
            elif str_op == "load_json":
                import json

                result = JacValue(self, value=json.loads(atom_res.value))
            elif len(kid) < 2 and str_op == "split":
                result = JacValue(self, value=atom_res.value.split())
            elif len(kid) < 2 and str_op == "strip":
                result = JacValue(self, value=atom_res.value.strip())
            elif len(kid) < 2 and str_op == "lstrip":
                result = JacValue(self, value=atom_res.value.lstrip())
            elif len(kid) < 2 and str_op == "rstrip":
                result = JacValue(self, value=atom_res.value.rstrip())
            if result:
                if len(kid) > 1:
                    self.rt_warn(f"{str_op} does not take parameters, ignoring", kid[2])
                return result
            if len(kid) > 1:
                args = self.run_expr_list(kid[2]).value
                if str_op == "count":
                    result = JacValue(self, value=atom_res.value.count(*args))
                elif str_op == "find":
                    result = JacValue(self, value=atom_res.value.find(*args))
                elif str_op == "split":
                    result = JacValue(self, value=atom_res.value.split(*args))
                elif str_op == "join":
                    if len(args) == 1 and type(args[0]) is list:
                        args = args[0]

                    result = JacValue(
                        self, value=atom_res.value.join([str(arg) for arg in args])
                    )
                elif str_op == "startswith":
                    result = JacValue(self, value=atom_res.value.startswith(*args))
                elif str_op == "endswith":
                    result = JacValue(self, value=atom_res.value.endswith(*args))
                elif str_op == "replace":
                    result = JacValue(self, value=atom_res.value.replace(*args))
                elif str_op == "strip":
                    result = JacValue(self, value=atom_res.value.strip(*args))
                elif str_op == "lstrip":
                    result = JacValue(self, value=atom_res.value.lstrip(*args))
                elif str_op == "rstrip":
                    result = JacValue(self, value=atom_res.value.rstrip(*args))
                if result:
                    return result
        except Exception as e:
            self.rt_error(f"{e}", jac_ast)
        self.rt_error(f"Call to {str_op} is invalid.", jac_ast)
        return atom_res

    def run_node_edge_ref(self, jac_ast):
        """
        node_edge_ref:
            node_ref filter_ctx?
            | edge_ref (node_ref filter_ctx?)?;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "node_ref":
            result = self.run_node_ref(kid[0])
            if len(kid) > 1:
                result = self.run_filter_ctx(kid[1], result)
            return JacValue(self, value=result)

        elif kid[0].name == "edge_ref":
            relevant_edges = self.run_edge_ref(kid[0])
            result = self.edge_to_node_jac_set(self.run_edge_ref(kid[0]))
            if len(kid) > 1 and kid[1].name == "node_ref":
                nres = self.run_node_ref(kid[1])
                if len(kid) > 2:
                    nres = self.run_filter_ctx(kid[2], nres)
                result = result * nres
                relevant_edges = self.edges_filter_on_nodes(relevant_edges, result)
            self._relevant_edges = relevant_edges
            return JacValue(self, value=result)

    def run_node_ref(self, jac_ast, is_spawn=False):
        """
        node_ref: KW_NODE DBL_COLON NAME;
        """
        kid = self.set_cur_ast(jac_ast)
        if not is_spawn:
            result = JacSet()
            if len(kid) > 1:
                for i in self.viable_nodes().obj_list():
                    if i.get_architype().is_instance(kid[2].token_text()):
                        result.add_obj(i)
            else:
                result += self.viable_nodes()
        else:
            result = self.parent().run_architype(
                kid[2].token_text(), kind="node", caller=self
            )
        return result

    def run_walker_ref(self, jac_ast, to_await=False):
        """
        walker_ref: KW_WALKER DBL_COLON NAME;
        """
        kid = self.set_cur_ast(jac_ast)
        name = kid[2].token_text()
        wlk = self.yielded_walkers_ids.get_obj_by_name(name, silent=True)
        if wlk is None:
            wlk = self.parent().run_architype(name=name, kind="walker", caller=self)
        if wlk is None:
            self.rt_error(f"No walker {name} exists!", kid[2])
        wlk._to_await = to_await
        return wlk

    def run_graph_ref(self, jac_ast):
        """
        graph_ref: KW_GRAPH DBL_COLON NAME;
        """
        kid = self.set_cur_ast(jac_ast)
        gph = self.parent().run_architype(
            kid[2].token_text(), kind="graph", caller=self
        )
        return gph

    def run_type_ref(self, jac_ast):
        """
        graph_ref: KW_GRAPH DBL_COLON NAME;
        """
        kid = self.set_cur_ast(jac_ast)
        obj = self.parent().run_architype(kid[2].token_text(), kind="type", caller=self)
        return JacValue(self, value=obj)

    def run_edge_ref(self, jac_ast, is_spawn=False):
        """
        edge_ref: edge_to | edge_from | edge_any;
        """
        kid = self.set_cur_ast(jac_ast)
        if not is_spawn:
            return self.run_rule(kid[0])
        else:
            if len(kid[0].kid) > 2:
                result = self.parent().run_architype(
                    kid[0].kid[2].token_text(), kind="edge", caller=self
                )
                if kid[0].kid[3].name == "spawn_ctx":
                    self.run_spawn_ctx(kid[0].kid[3], result)
                elif kid[0].kid[3].name == "filter_ctx":
                    self.rt_error("Filtering not allowed here", kid[0].kid[3])
            else:
                result = Edge(
                    m_id=self._m_id,
                    h=self._h,
                    kind="edge",
                    name="generic",
                )
            return result

    def run_edge_to(self, jac_ast):
        """
        edge_to:
            '-->'
            | '-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '->';
        """
        kid = self.set_cur_ast(jac_ast)
        result = JacSet()
        for i in (
            self.current_node.outbound_edges() + self.current_node.bidirected_edges()
        ):
            if len(kid) > 2 and not i.get_architype().is_instance(kid[2].token_text()):
                continue
            result.add_obj(i)
        if len(kid) > 2 and kid[3].name == "filter_ctx":
            result = self.run_filter_ctx(kid[3], result)
        elif len(kid) > 2 and kid[3].name == "spawn_ctx":
            self.rt_error("Assigning values not allowed here", kid[3])
        return result

    def run_edge_from(self, jac_ast):
        """
        edge_from:
            '<--'
            | '<-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '-';
        """
        kid = self.set_cur_ast(jac_ast)
        result = JacSet()
        for i in (
            self.current_node.inbound_edges() + self.current_node.bidirected_edges()
        ):
            if len(kid) > 2 and not i.get_architype().is_instance(kid[2].token_text()):
                continue
            result.add_obj(i)
        if len(kid) > 2 and kid[3].name == "filter_ctx":
            result = self.run_filter_ctx(kid[3], result)
        elif len(kid) > 2 and kid[3].name == "spawn_ctx":
            self.rt_error("Assigning values not allowed here", kid[3])
        return result

    def run_edge_any(self, jac_ast):
        """
        edge_any:
            '<-->'
            | '<-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '->';
        NOTE: these do not use strict bidirected semantic but any edge
        """
        kid = self.set_cur_ast(jac_ast)
        result = JacSet()
        for i in self.current_node.attached_edges():
            if len(kid) > 2 and not i.get_architype().is_instance(kid[2].token_text()):
                continue
            result.add_obj(i)
        if len(kid) > 2 and kid[3].name == "filter_ctx":
            result = self.run_filter_ctx(kid[3], result)
        elif len(kid) > 2 and kid[3].name == "spawn_ctx":
            self.rt_error("Assigning values not allowed here", kid[3])
        return result

    def run_list_val(self, jac_ast):
        """
        list_val: LSQUARE expr_list? RSQUARE;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[1].name == "expr_list":
            return self.run_expr_list(kid[1])
        return JacValue(self, value=[])

    def run_index_slice(self, jac_ast, atom_res):
        """
        index_slice:
            LSQUARE expression RSQUARE
            | LSQUARE expression COLON expression RSQUARE;
        """
        kid = self.set_cur_ast(jac_ast)
        idx = self.run_expression(kid[1]).value
        if (
            type(idx) == str
            and idx not in atom_res.value.keys()
            and not self._assign_mode
        ):
            self.rt_error(f"Key {idx} not found in object/dict.", kid[1])
        if kid[2].name == "RSQUARE":
            if not self.rt_check_type(idx, [int, str], kid[1]):
                self.rt_error(
                    f"Index of type {type(idx)} not valid. "
                    f"Indicies must be an integer or string!",
                    kid[1],
                )
                return atom_res
            atom_res.unwrap()
            try:
                return JacValue(self, ctx=atom_res.value, name=idx)
            except Exception:
                self.rt_error("List index out of range", kid[1])
                return atom_res
        else:
            end = self.run_expression(kid[3]).value
            if not self.rt_check_type(idx, [int], kid[1]) or not self.rt_check_type(
                end, [int], kid[3]
            ):
                self.rt_error(
                    "List slice range not valid. " "Indicies must be an integers!",
                    kid[1],
                )
                return atom_res
            atom_res.unwrap()
            try:
                return JacValue(self, ctx=atom_res.value, name=idx, end=end)
            except Exception:
                self.rt_error("List slice out of range", kid[1])
                return atom_res

    def run_dict_val(self, jac_ast):
        """
        dict_val: LBRACE (kv_pair (COMMA kv_pair)*)? RBRACE;
        """
        kid = self.set_cur_ast(jac_ast)
        dict_res = {}
        for i in kid:
            if i.name == "kv_pair":
                self.run_kv_pair(i, dict_res)
        return JacValue(self, value=dict_res)

    def run_kv_pair(self, jac_ast, obj):
        """
        kv_pair: STRING COLON expression;
        """
        kid = self.set_cur_ast(jac_ast)
        key = self.run_expression(kid[0]).value
        if isinstance(key, str):
            obj[key] = self.run_expression(kid[2]).value
        else:
            self.rt_error(f"Key is not str type : {type(key)}!", kid[0])

    def run_spawn(self, jac_ast):
        """
        spawn: KW_SPAWN spawn_object;

        NOTE: spawn statements support locations that are either nodes or
        jac_sets
        """
        kid = self.set_cur_ast(jac_ast)
        return self.run_spawn_object(kid[1])

    def run_spawn_object(self, jac_ast):
        """
        spawn_object: node_spawn | walker_spawn | graph_spawn;
        """
        kid = self.set_cur_ast(jac_ast)
        return self.run_rule(kid[0])

    def run_spawn_edge(self, jac_ast):
        """
        spawn_edge: expression edge_ref;
        """
        kid = self.set_cur_ast(jac_ast)
        loc = self.run_expression(kid[0]).value
        if isinstance(loc, JacSet):
            edge_set = [self.run_edge_ref(kid[1], is_spawn=True) for _ in loc]
            loc = loc.obj_list()
        else:
            edge_set = self.run_edge_ref(kid[1], is_spawn=True)
        return {
            "location": loc,
            "use_edge": edge_set,
            "direction": kid[1].kid[0].name,
        }

    def run_node_spawn(self, jac_ast):
        """
        node_spawn: spawn_edge? node_ref spawn_ctx?;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "node_ref":
            ret_node = self.run_node_ref(kid[0], is_spawn=True)
            if kid[-1].name == "spawn_ctx":
                self.run_spawn_ctx(kid[-1], ret_node)
            return JacValue(self, value=ret_node)
        else:
            sp = self.run_spawn_edge(kid[0])
            if isinstance(sp["location"], Node):
                ret_node = self.run_node_ref(kid[1], is_spawn=True)
                if sp["direction"] == "edge_from":
                    sp["location"].attach_inbound(ret_node, [sp["use_edge"]])
                elif sp["direction"] == "edge_to":
                    sp["location"].attach_outbound(ret_node, [sp["use_edge"]])
                else:
                    sp["location"].attach_bidirected(ret_node, [sp["use_edge"]])
                if kid[-1].name == "spawn_ctx":
                    self.run_spawn_ctx(kid[-1], ret_node)
                return JacValue(self, value=ret_node)
            elif isinstance(sp["location"], JacSet):
                res = []
                sp["location"] = sp["location"].obj_list()
                for i in range(len(sp["location"])):
                    ret_node = self.run_node_ref(kid[1], is_spawn=True)
                    if sp["direction"] == "edge_from":
                        sp["location"][i].attach_inbound(ret_node, [sp["use_edge"][i]])
                    elif sp["direction"] == "edge_to":
                        sp["location"][i].attach_outbound(ret_node, [sp["use_edge"][i]])
                    else:
                        sp["location"][i].attach_bidirected(
                            ret_node, [sp["use_edge"][i]]
                        )
                    if kid[-1].name == "spawn_ctx":
                        self.run_spawn_ctx(kid[-1], ret_node)
                    res.append(ret_node)
                return JacValue(self, value=res)
            else:
                self.rt_error(f'Spawn can not occur on {type(sp["location"])}!', kid[1])

    def run_graph_spawn(self, jac_ast):
        """
        graph_spawn: spawn_edge? graph_ref;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "graph_ref":
            result = self.run_graph_ref(kid[0])
            return JacValue(self, value=result)
        else:
            sp = self.run_spawn_edge(kid[0])
            if isinstance(sp["location"], Node):
                ret_node = self.run_graph_ref(kid[1])
                if sp["direction"] == "edge_from":
                    sp["location"].attach_inbound(ret_node, [sp["use_edge"]])
                elif sp["direction"] == "edge_to":
                    sp["location"].attach_outbound(ret_node, [sp["use_edge"]])
                else:
                    sp["location"].attach_bidirected(ret_node, [sp["use_edge"]])
                return JacValue(self, value=ret_node)
            elif isinstance(sp["location"], JacSet):
                res = []
                sp["location"] = sp["location"].obj_list()
                for i in range(len(sp["location"])):
                    ret_node = self.run_graph_ref(kid[1])
                    if sp["direction"] == "edge_from":
                        sp["location"][i].attach_inbound(ret_node, [sp["use_edge"][i]])
                    elif sp["direction"] == "edge_to":
                        sp["location"][i].attach_outbound(ret_node, [sp["use_edge"][i]])
                    else:
                        sp["location"][i].attach_bidirected(
                            ret_node, [sp["use_edge"][i]]
                        )
                    res.append(ret_node)
                return JacValue(self, value=res)
            else:
                self.rt_error(f'Spawn can not occur on {type(sp["location"])}!', kid[1])

    def run_walker_spawn(self, jac_ast):
        """
        walker_spawn: expression KW_SYNC? walker_ref spawn_ctx?;
        """
        kid = self.set_cur_ast(jac_ast)
        is_await = kid[1].name == "KW_SYNC" and bool(kid.pop(1))
        location = self.run_expression(kid[0]).value
        if isinstance(location, Node):
            location = JacSet(in_list=[location])
        ret = []
        for i in location.obj_list():
            walk = self.run_walker_ref(kid[1], is_await)
            walk.prime(i, request_ctx=getattr(self, "request_context", {}))
            if len(kid) > 2:
                self.run_spawn_ctx(kid[2], walk)

            res = walk.run()

            if walk.for_queue() and not res["is_queued"]:
                res["result"] = walk.anchor_value()

            tr = JacValue(self, value=res if walk.for_queue() else walk.anchor_value())

            tr.unwrap()
            ret.append(tr.value)
            self.inherit_runtime_state(walk)
            walk.register_yield_or_destroy(self.yielded_walkers_ids)
        return JacValue(self, value=ret[0] if len(ret) == 1 else ret)

    def run_spawn_ctx(self, jac_ast, obj):
        """
        spawn_ctx: LPAREN (spawn_assign (COMMA spawn_assign)*)? RPAREN;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "spawn_assign":
                self.run_spawn_assign(i, obj)

    def run_filter_ctx(self, jac_ast, obj):
        """
        filter_ctx:
                LPAREN (filter_compare (COMMA filter_compare)*)? RPAREN;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = JacSet()
        for i in obj.obj_list():
            for j in kid:
                if j.name == "filter_compare":
                    if self.run_filter_compare(j, i):
                        ret.add_obj(i)
        return ret

    def run_spawn_assign(self, jac_ast, obj):
        """
        spawn_assign: NAME EQ expression;
        """
        kid = self.set_cur_ast(jac_ast)
        name = kid[0].token_text()
        result = self.run_expression(kid[-1]).value
        if isinstance(obj, dict):
            obj[name] = result
            return
        dest = JacValue(self, ctx=obj, name=name, value=result)
        if obj.j_type == "walker":
            dest.write(kid[0], force=True)
        else:
            dest.write(kid[0])

    def run_filter_compare(self, jac_ast, obj):
        """
        filter_compare: NAME cmp_op expression;
        """
        kid = self.set_cur_ast(jac_ast)
        name = kid[0].token_text()
        if name in obj.context.keys():
            result = self.run_expression(kid[-1])
            return self.run_cmp_op(
                kid[1], JacValue(self, ctx=obj, name=name), result
            ).value
        else:
            self.rt_error(f"{name} not present in object", kid[0])
            return False

    def run_any_type(self, jac_ast):
        """
        any_type:
            TYP_STRING
            | TYP_INT
            | TYP_FLOAT
            | TYP_LIST
            | TYP_DICT
            | TYP_BOOL
            | KW_NODE
            | KW_EDGE
            | KW_TYPE;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "TYP_STRING":
            return JacValue(self, value=str)
        elif kid[0].name == "TYP_INT":
            return JacValue(self, value=int)
        elif kid[0].name == "TYP_FLOAT":
            return JacValue(self, value=float)
        elif kid[0].name == "TYP_LIST":
            return JacValue(self, value=list)
        elif kid[0].name == "TYP_DICT":
            return JacValue(self, value=dict)
        elif kid[0].name == "TYP_BOOL":
            return JacValue(self, value=bool)
        elif kid[0].name == "KW_NODE":
            return JacValue(self, value=Node)
        elif kid[0].name == "KW_EDGE":
            return JacValue(self, value=Edge)
        elif kid[0].name == "KW_TYPE":
            return JacValue(self, value=type)
        else:
            self.rt_error("Unrecognized type", kid[0])

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        super().destroy()

    # Helper Functions ##################

    def call_ability(self, nd, name, act_list):
        m = Interp(parent_override=self.parent(), caller=self)
        m.current_node = nd
        arch = nd.get_architype()
        m.push_scope(
            JacScope(parent=self, has_obj=nd, action_sets=[arch.get_all_actions()])
        )
        m._jac_scope.inherit_agent_refs(self._jac_scope, nd)
        try:
            m.run_code_block(jac_ir_to_ast(act_list.get_obj_by_name(name).value))
        except Exception as e:
            self.rt_error(f"Internal Exception: {e}", m._cur_jac_ast)
        self.inherit_runtime_state(m)

    def run_rule(self, jac_ast, *args):
        """Helper to run rule if exists in execution context"""
        try:
            return getattr(self, f"run_{jac_ast.name}")(jac_ast, *args)
        except AttributeError as e:
            if not hasattr(self, f"run_{jac_ast.name}"):
                self.rt_error(
                    f"This scope cannot execute the statement "
                    f'"{jac_ast.get_text()}" of type {jac_ast.name}',
                    jac_ast,
                )
            else:
                self.rt_error(f"{e}", jac_ast)
            return



from unittest import TestCase

import jaseci.actions.tests.std_test_code as stc
from jaseci.actor.sentinel import Sentinel
from jaseci.graph.graph import Graph
from jaseci.svc import MetaService
from jaseci.utils.utils import TestCaseHelper


class JacTests(TestCaseHelper, TestCase):
    """Unit tests for STD library language"""

    def setUp(self):
        super().setUp()
        self.meta = MetaService()

    def tearDown(self):
        super().tearDown()

    def test_rand_std(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(stc.rand_std)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(len(report), 4)
        self.assertGreater(len(report[1]), len(report[0]))
        self.assertGreater(len(report[2]), len(report[1]))
        self.assertGreater(len(report[3]), len(report[2]))

    def test_file_io(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(stc.file_io)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, ['{"a": 10}{"a": 10}'])

    def test_std_used_in_node_has_var(self):
        sent = Sentinel(m_id="anon", h=self.meta.build_hook())
        gph = Graph(m_id="anon", h=sent._h)
        sent.register_code(stc.std_used_in_node_has_var)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertGreater(len(report[0]), 10)
        self.assertEqual(type(report[0]), str)
